import numpy as np
import pytest

import starfish.data
from starfish import FieldOfView
from starfish.image import Filter
from starfish.spots import FindSpots
from starfish.types import TraceBuildingStrategies


@pytest.mark.skip('This test runs but takes forever')
def test_allen_smFISH_cropped_data():

    # set random seed to errors provoked by optimization functions
    np.random.seed(777)

    # load the experiment
    experiment = starfish.data.allen_smFISH(use_test_data=True)

    primary_image = experiment.fov().get_image(FieldOfView.PRIMARY_IMAGES)

    clip = Filter.Clip(p_min=10, p_max=100)
    clipped_image = clip.run(primary_image, in_place=False)

    bandpass = Filter.Bandpass(lshort=0.5, llong=7, threshold=0.0, truncate=4)
    bandpassed_image = bandpass.run(clipped_image, in_place=False)

    clip = Filter.Clip(p_min=10, p_max=100, is_volume=False)
    clipped_bandpassed_image = clip.run(bandpassed_image, in_place=False)

    sigma = (1, 0, 0)  # filter only in z, do nothing in x, y
    glp = Filter.GaussianLowPass(sigma=sigma, is_volume=True)
    z_filtered_image = glp.run(clipped_bandpassed_image, in_place=False)

    tlmpf = FindSpots.TrackpyLocalMaxPeakFinder(
        spot_diameter=5,  # must be odd integer
        min_mass=0.02,
        max_size=2,  # this is max radius
        separation=7,
        noise_size=0.65,  # this is not used because preprocess is False
        preprocess=False,
        percentile=10,
        # this is irrelevant when min_mass, spot_diameter, and max_size are set properly
        verbose=True,
        is_volume=True,
    )
    spots = tlmpf.run(z_filtered_image)  # noqa

    decoder = starfish.spots.DecodeSpots.PerRoundMaxChannel(
        codebook=experiment.codebook,
        trace_building_strategy=TraceBuildingStrategies.SEQUENTIAL
    )
    decoder.run(spots=spots)



from typing import Tuple


def _get_physical_coordinates_of_z_plane(zrange: Tuple[float, float]):
    """Calculate the midpoint of the given zrange."""
    physical_z = (zrange[1] - zrange[0]) / 2 + zrange[0]
    return physical_z



# Definition of the processing class
from functools import partial
from typing import Callable, Optional, Tuple, Union

import xarray as xr
from scipy.ndimage import gaussian_laplace

from starfish.core.image.Filter._base import FilterAlgorithm
from starfish.core.image.Filter.util import (
    determine_axes_to_group_by,
    validate_and_broadcast_kernel_size,
)
from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Clip, Levels, Number


class Laplace(FilterAlgorithm):
    """
    Multi-dimensional Gaussian-Laplacian filter used to enhance dots against background

    This filter wraps :py:func:`scipy.ndimage.gaussian_laplace`

    Parameters
    ----------
    sigma : Union[Number, Tuple[Number]]
        Standard deviation for Gaussian kernel to enhance dots.
    mode : str
        The mode parameter determines how the input array is extended when
        the filter overlaps a border. By passing a sequence of modes with
        length equal to the number of dimensions of the input array,
        different modes can be specified along each axis. Default value
        is ‘reflect’.
        The valid values and their behavior is as follows:

        ‘reflect’ (d c b a | a b c d | d c b a)
        The input is extended by reflecting about the edge of the last pixel.

        ‘constant’ (k k k k | a b c d | k k k k)
        The input is extended by filling all values beyond the edge with the same
        constant value, defined by the cval parameter.

        ‘nearest’ (a a a a | a b c d | d d d d)
        The input is extended by replicating the last pixel.

        ‘mirror’ (d c b | a b c d | c b a)
        The input is extended by reflecting about the center of the last pixel.

        ‘wrap’ (a b c d | a b c d | a b c d)
        The input is extended by wrapping around to the opposite edge.
    cval : scalar, optional
        (Default 0) Value to fill past edges of input if mode is ‘constant’.
    is_volume: bool
        If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) planes
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.
    """

    def __init__(
            self,
            sigma: Union[Number, Tuple[Number]],
            mode: str = 'reflect',
            cval: float = 0.0,
            is_volume: bool = False,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None
    ) -> None:

        self.sigma = validate_and_broadcast_kernel_size(sigma, is_volume=is_volume)
        self.mode = mode
        self.cval = cval
        self.is_volume = is_volume
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)

    _DEFAULT_TESTING_PARAMETERS = {"sigma": 0.5}

    @staticmethod
    def _gaussian_laplace(
        image: xr.DataArray, sigma: Union[Number, Tuple[Number]],
        mode: str = 'reflect', cval: float = 0.0
    ) -> xr.DataArray:
        filtered = gaussian_laplace(
            image, sigma=sigma, mode=mode, cval=cval)

        filtered = -filtered  # the peaks are negative so invert the signal

        return filtered

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        apply_filtering: Callable = partial(self._gaussian_laplace, sigma=self.sigma)
        return stack.apply(
            apply_filtering,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
            level_method=self.level_method,
        )



from starfish.core.experiment.builder import build_image, TileFetcher
from starfish.core.experiment.builder.defaultproviders import OnesTile, tile_fetcher_factory
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Axes


def synthetic_stack(
        num_round: int = 4,
        num_ch: int = 4,
        num_z: int = 12,
        tile_height: int = 50,
        tile_width: int = 40,
        tile_fetcher: TileFetcher = None,
) -> ImageStack:
    """generate a synthetic ImageStack

    Returns
    -------
    ImageStack :
        imagestack containing a tensor whose default shape is (2, 3, 4, 30, 20)
        and whose default values are all 1.

    """
    if tile_fetcher is None:
        tile_fetcher = tile_fetcher_factory(
            OnesTile,
            False,
            {Axes.Y: tile_height, Axes.X: tile_width},
        )

    collection = build_image(
        range(1),
        range(num_round),
        range(num_ch),
        range(num_z),
        tile_fetcher,
    )
    tileset = list(collection.all_tilesets())[0][1]

    return ImageStack.from_tileset(tileset)



from functools import partial
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
from skimage.feature import blob_dog, blob_doh, blob_log

from starfish.core.image.Filter.util import determine_axes_to_group_by
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.spots.FindSpots import spot_finding_utils
from starfish.core.types import (
    Axes,
    Features,
    Number,
    PerImageSliceSpotResults,
    SpotAttributes,
    SpotFindingResults,
)
from ._base import FindSpotsAlgorithm

blob_detectors = {
    'blob_dog': blob_dog,
    'blob_doh': blob_doh,
    'blob_log': blob_log
}


class BlobDetector(FindSpotsAlgorithm):
    """
    Multi-dimensional gaussian spot detector

    This method is a wrapper for skimage.feature.blob_log

    Parameters
    ----------
    min_sigma : Number
        The minimum standard deviation for Gaussian Kernel. Keep this low to
        detect smaller blobs.
    max_sigma : Number
        The maximum standard deviation for Gaussian Kernel. Keep this high to
        detect larger blobs.
    num_sigma : int
        The number of intermediate values of standard deviations to consider
        between `min_sigma` and `max_sigma`.
    threshold : float
        The absolute lower bound for scale space maxima. Local maxima smaller
        than threshold are ignored. Reduce this to detect blobs with less
        intensities.
    is_volume: bool
        If True, pass 3d volumes (x, y, z) to func, else pass 2d tiles (x, y) to func. (default:
        True)
    overlap : float [0, 1]
        If two spots have more than this fraction of overlap, the spots are combined
        (default: 0.5)
    measurement_type : str ['max', 'mean']
        name of the function used to calculate the intensity for each identified spot area
        (default: max)
    detector_method: str ['blob_dog', 'blob_doh', 'blob_log']
        name of the type of detection method used from skimage.feature (default: blob_log)

    Notes
    -----
    see also: http://scikit-image.org/docs/dev/auto_examples/features_detection/plot_blob.html

    """

    def __init__(
            self,
            min_sigma: Union[Number, Tuple[Number, ...]],
            max_sigma: Union[Number, Tuple[Number, ...]],
            num_sigma: int,
            threshold: Number,
            overlap: float = 0.5,
            measurement_type='max',
            is_volume: bool = True,
            detector_method: str = 'blob_log',
            exclude_border: Optional[int] = None,
    ) -> None:

        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.num_sigma = num_sigma
        self.threshold = threshold
        self.overlap = overlap
        self.is_volume = is_volume
        self.measurement_function = self._get_measurement_function(measurement_type)
        self.exclude_border = exclude_border
        try:
            self.detector_method = blob_detectors[detector_method]
        except ValueError:
            raise ValueError("Detector method must be one of {blob_log, blob_dog, blob_doh}")

    def image_to_spots(
            self, data_image: np.ndarray,
    ) -> PerImageSliceSpotResults:
        """
        Find spots using a gaussian blob finding algorithm

        Parameters
        ----------
        data_image : np.ndarray
            image containing spots to be detected

        Returns
        -------
        PerImageSpotResults :
            includes a SpotAttributes DataFrame of metadata containing the coordinates, intensity
            and radius of each spot, as well as any extra information collected during spot finding.

        """

        fitted_blobs_array: np.ndarray = self.detector_method(
            data_image,
            min_sigma=self.min_sigma,
            max_sigma=self.max_sigma,
            threshold=self.threshold,
            exclude_border=self.exclude_border,
            overlap=self.overlap,
            num_sigma=self.num_sigma
        )

        if fitted_blobs_array.shape[0] == 0:
            empty_spot_attrs = SpotAttributes.empty(
                extra_fields=[Features.INTENSITY, Features.SPOT_ID])
            return PerImageSliceSpotResults(spot_attrs=empty_spot_attrs, extras=None)

        # measure intensities
        z_inds = fitted_blobs_array[:, 0].astype(int)
        y_inds = fitted_blobs_array[:, 1].astype(int)
        x_inds = fitted_blobs_array[:, 2].astype(int)
        radius = np.round(fitted_blobs_array[:, 3] * np.sqrt(3))
        data_image = np.asarray(data_image)
        intensities = data_image[tuple([z_inds, y_inds, x_inds])]

        # construct dataframe
        spot_data = pd.DataFrame(
            data={
                Features.INTENSITY: intensities,
                Axes.ZPLANE.value: z_inds,
                Axes.Y.value: y_inds,
                Axes.X.value: x_inds,
                Features.SPOT_RADIUS: radius,
            }
        )
        spots = SpotAttributes(spot_data)
        spots.data[Features.SPOT_ID] = np.arange(spots.data.shape[0])
        return PerImageSliceSpotResults(spot_attrs=spots, extras=None)

    def run(
            self,
            image_stack: ImageStack,
            reference_image: Optional[ImageStack] = None,
            n_processes: Optional[int] = None,
            *args,
    ) -> SpotFindingResults:
        """
        Find spots in the given ImageStack using a gaussian blob finding algorithm.
        If a reference image is provided the spots will be detected there then measured
        across all rounds and channels in the corresponding ImageStack. If a reference_image
        is not provided spots will be detected _independently_ in each channel. This assumes
        a non-multiplex imaging experiment, as only one (ch, round) will be measured for each spot.

        Parameters
        ----------
        image_stack : ImageStack
            ImageStack where we find the spots in.
        reference_image : Optional[ImageStack]
            (Optional) a reference image. If provided, spots will be found in this image, and then
            the locations that correspond to these spots will be measured across each channel.
        n_processes : Optional[int] = None,
            Number of processes to devote to spot finding.
        """
        spot_finding_method = partial(self.image_to_spots, *args)
        if reference_image:
            data_image = reference_image._squeezed_numpy(*{Axes.ROUND, Axes.CH})
            reference_spots = spot_finding_method(data_image)
            results = spot_finding_utils.measure_intensities_at_spot_locations_across_imagestack(
                data_image=image_stack,
                reference_spots=reference_spots,
                measurement_function=self.measurement_function)
        else:
            spot_attributes_list = image_stack.transform(
                func=spot_finding_method,
                group_by=determine_axes_to_group_by(self.is_volume),
                n_processes=n_processes
            )
            results = SpotFindingResults(imagestack_coords=image_stack.xarray.coords,
                                         log=image_stack.log,
                                         spot_attributes_list=spot_attributes_list)
        return results



import csv
import dataclasses
import os
from pathlib import Path
from typing import Callable, Mapping, MutableMapping, MutableSequence, Optional, Sequence

import numpy as np
from slicedimage import ImageFormat

from starfish.core.types import Axes, Coordinates, CoordinateValue
from ..factories import unique_data
from ... import FetchedTile, TileFetcher, TileIdentifier
from ...structured_formatter import TILE_COORDINATE_NAMES


class UniqueTiles(FetchedTile):
    """Tiles where the pixel values are unique per fov/round/ch/z."""
    def __init__(
            self,
            # these are the arguments passed in as a result of tile_fetcher_factory's
            # pass_tile_indices parameter.
            fov_id: int, round_id: int, ch_id: int, zplane_id: int,
            # these are the arguments we are passing through tile_fetcher_factory.
            fovs: Sequence[int], rounds: Sequence[int], chs: Sequence[int], zplanes: Sequence[int],
            tile_height: int, tile_width: int,
    ) -> None:
        super().__init__()
        self.fov_id = fov_id
        self.round_id = round_id
        self.ch_id = ch_id
        self.zplane_id = zplane_id
        self.fovs = fovs
        self.rounds = rounds
        self.chs = chs
        self.zplanes = zplanes
        self.tile_height = tile_height
        self.tile_width = tile_width

    @property
    def shape(self) -> Mapping[Axes, int]:
        return {Axes.Y: self.tile_height, Axes.X: self.tile_width}

    def tile_data(self) -> np.ndarray:
        """Return the data for a given tile."""
        return unique_data(
            self.fov_id,
            self.rounds.index(self.round_id),
            self.chs.index(self.ch_id),
            self.zplanes.index(self.zplane_id),
            len(self.fovs),
            len(self.rounds),
            len(self.chs),
            len(self.zplanes),
            self.tile_height,
            self.tile_width)


def write_tile_data(
        basepath: Path,
        image_type: str,
        tile_format: ImageFormat,
        tile_identifiers: Sequence[TileIdentifier],
        fetcher: TileFetcher,
        tilepath_generator: Optional[Callable[
            [Path, str, ImageFormat, TileIdentifier], Path]] = None,
) -> None:
    if tilepath_generator is None:
        tilepath_generator = \
            lambda _basepath, _image_type, _tile_format, _tile_identifier: basepath / (
                f"{_image_type}-"
                f"f{_tile_identifier.fov_id}-"
                f"r{_tile_identifier.round_label}-"
                f"c{_tile_identifier.ch_label}-"
                f"z{_tile_identifier.zplane_label}."
                f"{tile_format.file_ext}"
            )

    for tile_identifier in tile_identifiers:
        fetched_tile = fetcher.get_tile(
            tile_identifier.fov_id,
            tile_identifier.round_label,
            tile_identifier.ch_label,
            tile_identifier.zplane_label)

        tilepath = tilepath_generator(basepath, image_type, tile_format, tile_identifier)
        tilepath.parent.mkdir(parents=True, exist_ok=True)

        tile_format.writer_func(tilepath, fetched_tile.tile_data())


def render_coordinates_to_rows(
        tile_to_physical_coordinates: Mapping[
            TileIdentifier, Mapping[Coordinates, CoordinateValue]],
) -> Sequence[Mapping[str, str]]:
    results: MutableSequence[Mapping[str, str]] = list()

    for tile_identifier, physical_coordinates in tile_to_physical_coordinates.items():
        rowdata: MutableMapping[str, str] = dict()

        for tile_coordinate_name, tile_coordinate_value in zip(
                TILE_COORDINATE_NAMES, dataclasses.astuple(tile_identifier)
        ):
            rowdata[tile_coordinate_name] = str(tile_coordinate_value)

        for coordinate_name in list(Coordinates):
            coordinate_value = physical_coordinates.get(coordinate_name, None)
            if coordinate_value is None and coordinate_name == Coordinates.Z:
                # Z coordinates may be legitimately missing
                continue

            if isinstance(coordinate_value, tuple):
                rowdata[f'{coordinate_name}_min'] = str(coordinate_value[0])
                rowdata[f'{coordinate_name}_max'] = str(coordinate_value[1])
            else:
                rowdata[f'{coordinate_name}_min'] = str(coordinate_value)

        results.append(rowdata)

    return results


def write_coordinates_csv(
        path: Path,
        rows: Sequence[Mapping[str, str]],
        write_z_coordinates_in_header: bool,
) -> None:
    headers = list(TILE_COORDINATE_NAMES)
    for coordinate_name in list(Coordinates):
        if coordinate_name == Coordinates.Z and not write_z_coordinates_in_header:
            continue
        headers.append(f"{coordinate_name.value}_min")
        headers.append(f"{coordinate_name.value}_max")

    with open(os.fspath(path), "w") as fh:
        writer = csv.DictWriter(fh, headers)
        writer.writeheader()
        for row in rows:
            writer.writerow(row)



import codecs
import io
import pickle
import tarfile
import warnings
from dataclasses import dataclass
from typing import BinaryIO, List, Mapping, MutableSequence, Optional, Tuple, Type

import numpy as np
from packaging.version import Version

from starfish.core.errors import DataFormatWarning
from starfish.core.types import ArrayLike, Axes, Coordinates, Number, STARFISH_EXTRAS_KEY
from starfish.core.util.logging import Log
from . import binary_mask as bm


class AttrKeys:
    DOCTYPE = f"{STARFISH_EXTRAS_KEY}.DOCTYPE"
    VERSION = f"{STARFISH_EXTRAS_KEY}.VERSION"


DOCTYPE_STRING = "starfish/BinaryMaskCollection"


class BinaryMaskIO:
    ASCENDING_VERSIONS: List[Tuple[Version, Type["BinaryMaskIO"]]] = []
    VERSION: Version

    def __init_subclass__(cls, version_descriptor: Version, **kwargs):
        super.__init_subclass__(**kwargs)  # type: ignore
        cls.VERSION = version_descriptor
        BinaryMaskIO.ASCENDING_VERSIONS.append((version_descriptor, cls))
        BinaryMaskIO.ASCENDING_VERSIONS.sort()

    @staticmethod
    def read_versioned_binary_mask(
            file_obj: BinaryIO,
    ) -> bm.BinaryMaskCollection:
        with tarfile.open(mode="r|gz", fileobj=file_obj) as tf:
            if tf.pax_headers is None:
                raise ValueError(
                    f"file does not appear to be a binary mask file (does not have headers)")
            if tf.pax_headers.get(AttrKeys.DOCTYPE, None) != DOCTYPE_STRING:
                raise ValueError(
                    f"file does not appear to be a binary mask file (missing or incorrect doctype)")
            if AttrKeys.VERSION not in tf.pax_headers:
                raise ValueError(f"file missing version data")

            requested_version = Version(tf.pax_headers[AttrKeys.VERSION])

            for version, implementation in BinaryMaskIO.ASCENDING_VERSIONS:
                if version == requested_version:
                    return implementation().read_binary_mask(tf)
            else:
                raise ValueError(f"No reader for version {requested_version}")

    @staticmethod
    def write_versioned_binary_mask(
            file_obj: BinaryIO,
            binary_mask: bm.BinaryMaskCollection,
            requested_version: Optional[Version] = None,
    ):
        if requested_version is None:
            implementation = BinaryMaskIO.ASCENDING_VERSIONS[-1][1]
        else:
            for version, implementing_class in BinaryMaskIO.ASCENDING_VERSIONS:
                if version == requested_version:
                    implementation = implementing_class
                    break
            else:
                raise ValueError(f"No writer for version {requested_version}")

        implementation().write_binary_mask(file_obj, binary_mask)

    def read_binary_mask(self, tf: tarfile.TarFile) -> bm.BinaryMaskCollection:
        raise NotImplementedError()

    def write_binary_mask(self, file_obj: BinaryIO, binary_mask: bm.BinaryMaskCollection):
        raise NotImplementedError()


class v0_0(BinaryMaskIO, version_descriptor=Version("0.0")):
    LOG_FILENAME = "log"
    PIXEL_TICKS_FILE = "pixel_ticks.pickle"
    PHYSICAL_TICKS_FILE = "physical_ticks.pickle"
    MASK_PREFIX = "masks/"

    @dataclass
    class MaskOnDisk:
        binary_mask: np.ndarray
        offsets: Tuple[int, ...]

    def read_binary_mask(self, tf: tarfile.TarFile) -> bm.BinaryMaskCollection:
        log: Optional[Log] = None
        masks: MutableSequence[bm.MaskData] = []
        pixel_ticks: Optional[Mapping[Axes, ArrayLike[int]]] = None
        physical_ticks: Optional[Mapping[Coordinates, ArrayLike[Number]]] = None

        while True:
            tarinfo: Optional[tarfile.TarInfo] = tf.next()
            if tarinfo is None:
                break

            # wrap it in a BytesIO object to ensure we never seek backwards.
            extracted_fh = tf.extractfile(tarinfo)
            if extracted_fh is None:
                raise ValueError(f"Unable to extract file {tarinfo.name}")
            byte_stream = io.BytesIO(extracted_fh.read())

            if tarinfo.name == v0_0.LOG_FILENAME:
                string_stream = codecs.getreader("utf-8")(byte_stream)
                log = Log.decode(string_stream.read())
            elif tarinfo.name == v0_0.PIXEL_TICKS_FILE:
                pixel_ticks = pickle.load(byte_stream)
            elif tarinfo.name == v0_0.PHYSICAL_TICKS_FILE:
                physical_ticks = pickle.load(byte_stream)
            elif tarinfo.name.startswith(v0_0.MASK_PREFIX):
                mask_on_disk: v0_0.MaskOnDisk = pickle.load(byte_stream)
                if not isinstance(mask_on_disk, v0_0.MaskOnDisk):
                    raise TypeError(f"mask does not conform to expected mask structure")
                masks.append(bm.MaskData(mask_on_disk.binary_mask, mask_on_disk.offsets, None))
            else:
                warnings.warn(
                    f"Unexpected file in binary mask collection {tarinfo.name}",
                    DataFormatWarning
                )

        if pixel_ticks is None:
            raise ValueError("pixel coordinates not found")
        if physical_ticks is None:
            raise ValueError("physical coordinates not found")

        return bm.BinaryMaskCollection(pixel_ticks, physical_ticks, masks, log)

    def write_binary_mask(self, file_obj: BinaryIO, binary_mask: bm.BinaryMaskCollection):
        pax_headers: Mapping[str, str] = {
            AttrKeys.DOCTYPE: DOCTYPE_STRING,
            AttrKeys.VERSION: str(v0_0.VERSION),
        }
        with tarfile.open(
                fileobj=file_obj,
                mode="w|gz",
                format=tarfile.PAX_FORMAT,
                pax_headers=pax_headers,
        ) as tf:
            # write the log
            log_bytes = binary_mask._log.encode().encode("utf-8")
            write_to_tarfile(tf, v0_0.LOG_FILENAME, log_bytes)

            # write the pixel ticks
            pixel_ticks_bytes = pickle.dumps(binary_mask._pixel_ticks)
            write_to_tarfile(tf, v0_0.PIXEL_TICKS_FILE, pixel_ticks_bytes)

            # write the physical coordinate ticks
            physical_ticks_bytes = pickle.dumps(binary_mask._physical_ticks)
            write_to_tarfile(tf, v0_0.PHYSICAL_TICKS_FILE, physical_ticks_bytes)

            for ix, mask in binary_mask._masks.items():
                mask_on_disk = v0_0.MaskOnDisk(mask.binary_mask, mask.offsets)
                mask_bytes = pickle.dumps(mask_on_disk)
                write_to_tarfile(tf, f"{v0_0.MASK_PREFIX}{ix}", mask_bytes)


def write_to_tarfile(tf: tarfile.TarFile, name: str, data: bytes):
    tarinfo = tarfile.TarInfo(name=name)
    tarinfo.size = len(data)
    tf.addfile(tarinfo, io.BytesIO(data))



from collections import defaultdict
from typing import Any, Dict, Hashable, List, Mapping, Sequence, Tuple

import numpy as np
import pandas as pd
from sklearn.neighbors import NearestNeighbors

from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import Axes, Features, SpotFindingResults


def _match_spots(
    round_dataframes: Dict[int, pd.DataFrame], anchor_round: int
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """ For each spot in anchor round, find the closest spot within search_radius in all rounds.

    Parameters
    ----------
    round_dataframes : Dict[int, pd.DataFrame]
        Output from _merge_spots_by_round, contains mapping of image volumes from each round to
        all the spots detected in them.
    anchor_round : int
        The imaging round to seed the local search from.

    Returns
    -------
    pd.DataFrame
        Spots x rounds dataframe containing the distances to the nearest spot. np.nan if
        no spot is detected within search radius
    """
    reference_df = round_dataframes[anchor_round]
    reference_coordinates = reference_df[[Axes.ZPLANE, Axes.Y, Axes.X]]

    dist = pd.DataFrame(
        data=np.zeros((reference_df.shape[0], len(round_dataframes)), dtype=float),
        columns=list(round_dataframes.keys())
    )
    ind = pd.DataFrame(
        data=np.zeros((reference_df.shape[0], len(round_dataframes)), dtype=np.int32),
        columns=list(round_dataframes.keys())
    )

    # fill data for anchor round; every spot is a perfect match to itself.
    ind[anchor_round] = np.arange(reference_df.shape[0], dtype=np.int32)

    # get spots matching across rounds
    for r in sorted(set(round_dataframes.keys()) - {anchor_round, }):
        query_df = round_dataframes[r]
        query_coordinates = query_df[[Axes.ZPLANE, Axes.Y, Axes.X]]
        nn = NearestNeighbors(n_neighbors=1)
        nn.fit(query_coordinates)
        distances, indices = nn.kneighbors(reference_coordinates)
        dist[r] = distances
        ind[r] = indices

    return dist, ind


def _build_intensity_table(
    round_dataframes: Dict[int, pd.DataFrame],
    dist: pd.DataFrame,
    ind: pd.DataFrame,
    channels: Sequence[int],
    rounds: Sequence[int],
    search_radius: int,
    anchor_round: int,
) -> IntensityTable:
    """Construct an intensity table from the results of a local search over detected spots

    Parameters
    ----------
    round_dataframes : Dict[int, pd.DataFrame]
        Output from _merge_spots_by_round, contains mapping of image volumes from each round to
        all the spots detected in them.
    dist, ind : pd.DataFrame
        Output from _match_spots, contains distances and indices to the nearest spot for each
        spot in anchor_round.
    channels, rounds : Sequence[int]
        Channels and rounds present in the ImageStack from which spots were detected.
    search_radius : int
        The maximum (euclidean) distance in pixels for a spot to be considered matching in
        a round subsequent to the anchor round.
    anchor_round : int
        The imaging round to seed the local search from.

    """

    anchor_df = round_dataframes[anchor_round]

    # create empty IntensityTable filled with np.nan
    data = np.full((dist.shape[0], len(channels), len(rounds)), fill_value=np.nan)
    dims = (Features.AXIS, Axes.CH.value, Axes.ROUND.value)
    coords: Mapping[Hashable, Tuple[str, Any]] = {
        Features.SPOT_RADIUS: (Features.AXIS, anchor_df[Features.SPOT_RADIUS]),
        Axes.ZPLANE.value: (Features.AXIS, anchor_df[Axes.ZPLANE]),
        Axes.Y.value: (Features.AXIS, anchor_df[Axes.Y]),
        Axes.X.value: (Features.AXIS, anchor_df[Axes.X]),
        Axes.ROUND.value: (Axes.ROUND.value, rounds),
        Axes.CH.value: (Axes.CH.value, channels)
    }
    intensity_table = IntensityTable(data=data, dims=dims, coords=coords)

    # fill IntensityTable
    for r in rounds:
        # get intensity data and indices
        spot_indices = ind[r]
        intensity_data = round_dataframes[r].loc[spot_indices, Features.INTENSITY]
        channel_index = round_dataframes[r].loc[spot_indices, Axes.CH]
        round_index = np.full(ind.shape[0], fill_value=r, dtype=int)
        feature_index = np.arange(ind.shape[0], dtype=int)

        # mask spots that are outside the search radius
        mask = np.asarray(dist[r] < search_radius)  # indices need not match
        feature_index = feature_index[mask]
        channel_index = channel_index[mask]
        round_index = round_index[mask]
        intensity_data = intensity_data[mask]

        # need numpy indexing to set values in vectorized manner
        intensity_table.values[feature_index, channel_index, round_index] = intensity_data

    return intensity_table


def _merge_spots_by_round(
    spot_results: SpotFindingResults
) -> Dict[int, pd.DataFrame]:
    """Merge DataFrames containing spots from different channels into one DataFrame per round.

    Parameters
    ----------
    spot_results : Dict[Tuple[int, int], pd.DataFrame]
        Output of _find_spots. Dictionary mapping (round, channel) volumes to the spots detected
        in them.

    Returns
    -------
    Dict[int, pd.DataFrame]
        Dictionary mapping round volumes to the spots detected in them. Contains an additional
        column labeled by Axes.CH which identifies the channel in which a given spot was
        detected.

    """

    # add channel information to each table and add it to round_data
    round_data: Mapping[int, List] = defaultdict(list)
    for (r, c), df in spot_results.items():
        df = df.spot_attrs.data
        df[Axes.CH.value] = np.full(df.shape[0], c)
        round_data[r].append(df)

    # create one dataframe per round
    round_dataframes = {
        k: pd.concat(v, axis=0).reset_index().drop('index', axis=1)
        for k, v in round_data.items()
    }

    return round_dataframes



from typing import List, Mapping, MutableMapping, Optional, Tuple, Union

import numpy as np

from starfish.core.types import ArrayLike, Axes, Coordinates, Number


def _get_axes_names(ndim: int) -> Tuple[List[Axes], List[Coordinates]]:
    """Get needed axes and coordinates given the number of dimensions.  The axes and coordinates are
    returned in the order expected for binary masks.  For instance, the first axis/coordinate
    should be the first index into the mask.

    Parameters
    ----------
    ndim : int
        Number of dimensions.
    Returns
    -------
    axes : List[Axes]
        Axes.
    coords : List[Coordinates]
        Coordinates.
    """
    if ndim == 2:
        axes = [Axes.Y, Axes.X]
        coords = [Coordinates.Y, Coordinates.X]
    elif ndim == 3:
        axes = [Axes.ZPLANE, Axes.Y, Axes.X]
        coords = [Coordinates.Z, Coordinates.Y, Coordinates.X]
    else:
        raise TypeError('expected 2- or 3-D image')

    return axes, coords


def _normalize_pixel_ticks(
        pixel_ticks: Optional[Union[
            Mapping[Axes, ArrayLike[int]],
            Mapping[str, ArrayLike[int]]]],
) -> MutableMapping[Axes, ArrayLike[int]]:
    """Given pixel ticks in a mapping from an axis or a string representing an axis, return a
    mapping from an axis.  The mapping may also not be present (i.e., None), in which an empty
    dictionary is returned.
    """
    return {
        Axes(axis): axis_data
        for axis, axis_data in (pixel_ticks or {}).items()
    }


def _normalize_physical_ticks(
        physical_ticks: Union[
            Mapping[Coordinates, ArrayLike[Number]],
            Mapping[str, ArrayLike[Number]]],
) -> Mapping[Coordinates, ArrayLike[Number]]:
    """Given physical coordinate ticks in a mapping from a coordinate or a string representing a
    coordinate, return a mapping from a coordinate.
    """
    return {
        Coordinates(coord): coord_data
        for coord, coord_data in physical_ticks.items()
    }


def _ticks_equal(
        left: Mapping,
        right: Mapping,
) -> bool:
    """Given two sets of tick marks, return True if the two contain the same keys, and contain the
    same data for each key.  This works for both pixel ticks and physical ticks.
    """
    if left.keys() != right.keys():
        return False
    for key in left.keys():
        if not np.all(left[key] == right[key]):
            return False

    return True



from abc import abstractmethod
from typing import Optional

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.pipeline.algorithmbase import AlgorithmBase


class FilterAlgorithm(metaclass=AlgorithmBase):

    @abstractmethod
    def run(self, stack: ImageStack, *args) -> Optional[ImageStack]:
        """Perform filtering of an image stack"""
        raise NotImplementedError()



from itertools import product
from json import loads
from typing import Dict, Hashable, List, Optional, Sequence

import numpy as np
import pandas as pd
import xarray as xr

from starfish.core.types import (
    Axes,
    Coordinates,
    Features,
    LOG,
    OverlapStrategy,
    SpotAttributes,
    STARFISH_EXTRAS_KEY
)
from starfish.core.util.levels import preserve_float_range
from .overlap import (
    find_overlaps_of_xarrays,
    OVERLAP_STRATEGY_MAP,
)


class IntensityTable(xr.DataArray):
    """
    IntensityTable is a container for spot or pixel features extracted from
    image data. It is the primary output from starfish :py:class:`SpotFinder`
    methods.

    An IntensityTable records the numeric intensity of a set of features in each
    :code:`(round, channel)` tile in which the feature is identified.
    The :py:class:`IntensityTable` has shape
    :code:`(n_feature, n_channel, n_round)`.

    Some :py:class:`SpotFinder` methods identify a position and search for
    Gaussian blobs in a small radius, only recording intensities if they are
    found in a given tile. Other :py:class:SpotFinder: approaches
    find blobs in a max-projection and measure them everywhere. As a result,
    some IntensityTables will be dense, and others will contain :code:`np.nan`
    entries where no feature was detected.

    Examples
    --------
    Create an IntensityTable using the ``SyntheticData`` factory::

        >>> from starfish.core.test.factories import SyntheticData
        >>> sd = SyntheticData(n_ch=3, n_round=4, n_codes=2, n_spots=3)
        >>> codes = sd.codebook()
        >>> sd.intensities(codebook=codes)
        <xarray.IntensityTable (features: 3, r: 4, c: 3)>
        array([[[1., 0., 0.],
                [0., 1., 0.],
                [0., 0., 1.],
                [0., 1., 0.]],

               [[1., 0., 0.],
                [0., 1., 0.],
                [0., 0., 1.],
                [0., 1., 0.]],

               [[1., 0., 0.],
                [0., 1., 0.],
                [0., 0., 1.],
                [0., 1., 0.]]], dtype=float32)
        Coordinates:
            z         (features) int64 8 6 2
            y         (features) float64 14.61 41.9 3.935
            x         (features) float64 11.0 42.42 5.249
            radius    (features) float64 nan nan nan
          * features  (features) int64 0 1 2
          * r         (r) int64 0 1 2 3
          * c         (c) int64 0 1 2
    """

    __slots__ = ()

    @staticmethod
    def _build_xarray_coords(
            spot_attributes: SpotAttributes,
            round_values: Sequence[int],
            channel_values: Sequence[int],
    ) -> Dict[Hashable, np.ndarray]:
        """build coordinates for intensity-table"""
        coordinates = {
            k: (Features.AXIS, spot_attributes.data[k].values)
            for k in spot_attributes.data}
        coordinates.update({
            Features.AXIS: np.arange(len(spot_attributes.data)),
            Axes.ROUND.value: np.array(round_values),
            Axes.CH.value: np.array(channel_values),
        })
        return coordinates

    @classmethod
    def zeros(
            cls,
            spot_attributes: SpotAttributes,
            round_labels: Sequence[int],
            ch_labels: Sequence[int],
    ) -> "IntensityTable":
        """
        Create an empty intensity table with pre-set shape whose values are zero.

        Parameters
        ----------
        spot_attributes : SpotAttributes
            Table containing spot metadata. Must contain the values specified in Axes.X,
            Y, Z, and RADIUS.
        round_labels : Sequence[int]
            The possible values for the round number, in the order that they are in the ImageStack
            5D tensor.
        ch_labels : Sequence[int]
            The possible values for the channel number, in the order that they are in the ImageStack
            5D tensor.

        Returns
        -------
        IntensityTable :
            IntensityTable filled with zeros.

        """
        if not isinstance(spot_attributes, SpotAttributes):
            raise TypeError('parameter spot_attributes must be a starfish SpotAttributes object.')

        data = np.zeros((spot_attributes.data.shape[0], len(round_labels), len(ch_labels)))
        dims = (Features.AXIS, Axes.ROUND.value, Axes.CH.value,)
        coords = cls._build_xarray_coords(spot_attributes, round_labels, ch_labels)

        intensity_table = cls(
            data=data, coords=coords, dims=dims,
        )

        return intensity_table

    @classmethod
    def from_spot_data(
            cls,
            intensities: xr.DataArray,
            spot_attributes: SpotAttributes,
            round_values: Sequence[int],
            ch_values: Sequence[int],
            *args, **kwargs) -> "IntensityTable":
        """
        Creates an IntensityTable from a :code:`(features, channel, round)`
        array and a :py:class:`~starfish.types._spot_attributes.SpotAttributes`
        object.

        Parameters
        ----------
        intensities : xr.DataArray
            Intensity data.
        spot_attributes : SpotAttributes
            Table containing spot metadata. Must contain the values specified in Axes.X,
            Y, Z, and RADIUS.
        ch_values : Sequence[int]
            The possible values for the channel number, in the order that they are in the ImageStack
            5D tensor.
        round_values : Sequence[int]
            The possible values for the round number, in the order that they are in the ImageStack
        args :
            Additional arguments to pass to the xarray constructor.
        kwargs :
            Additional keyword arguments to pass to the xarray constructor.

        Returns
        -------
        IntensityTable :
            IntensityTable containing data from passed intensities, annotated by spot_attributes
        """

        if len(intensities.shape) != 3:
            raise ValueError(
                f'intensities must be a (features * ch * round) 3-d tensor. Provided intensities '
                f'shape ({intensities.shape}) is invalid.')

        if len(round_values) != intensities.shape[1]:
            raise ValueError(
                f"The number of round values ({len(round_values)}) should be equal to intensities' "
                f"shape[1] ({intensities.shape[1]})."
            )

        if len(ch_values) != intensities.shape[2]:
            raise ValueError(
                f"The number of ch values ({len(ch_values)}) should be equal to intensities' "
                f"shape[2] ({intensities.shape[2]})."
            )

        if not isinstance(spot_attributes, SpotAttributes):
            raise TypeError('parameter spot_attributes must be a starfish SpotAttributes object.')

        coords = cls._build_xarray_coords(spot_attributes, round_values, ch_values)

        dims = (Features.AXIS, Axes.ROUND.value, Axes.CH.value)

        intensities = cls(intensities, coords, dims, *args, **kwargs)
        return intensities

    def get_log(self):
        """
        Deserialize and return a list of pipeline components that have been applied
        throughout a starfish session to create this :py:class:IntensityTable:.
        """

        if STARFISH_EXTRAS_KEY in self.attrs and LOG in self.attrs[STARFISH_EXTRAS_KEY]:
            return loads(self.attrs[STARFISH_EXTRAS_KEY])[LOG]
        else:
            raise RuntimeError('No log info found.')

    @property
    def has_physical_coords(self):
        """Returns True if this table's features have physical-space loci."""
        return Coordinates.X in self.coords and Coordinates.Y in self.coords

    def to_netcdf(self, filename: str) -> None:  # type: ignore
        """
        Save an IntensityTable as a Netcdf File.

        Parameters
        ----------
        filename : str
            Name of Netcdf file.

        """
        super().to_netcdf(filename)

    @classmethod
    def open_netcdf(cls, filename: str) -> "IntensityTable":
        """
        Load an IntensityTable from Netcdf.

        Parameters
        ----------
        filename : str
            File to load.

        Returns
        -------
        IntensityTable

        """
        loaded = xr.open_dataarray(filename)
        intensity_table = cls(
            loaded.data,
            loaded.coords,
            loaded.dims,
            attrs=loaded.attrs,
        )
        return intensity_table

    @classmethod
    def synthetic_intensities(
            cls,
            codebook,
            num_z: int = 12,
            height: int = 50,
            width: int = 40,
            n_spots: int = 10,
            mean_fluor_per_spot: int = 200,
            mean_photons_per_fluor: int = 50,
    ) -> "IntensityTable":
        """
        Creates an IntensityTable with synthetic spots, that correspond to valid
        codes in a provided codebook.

        Parameters
        ----------
        codebook : Codebook
            Starfish codebook object.
        num_z : int
            Number of z-planes to use when localizing spots.
        height : int
            y dimension of each synthetic plane.
        width : int
            x dimension of each synthetic plane.
        n_spots : int
            Number of spots to generate.
        mean_fluor_per_spot : int
            Mean number of fluorophores per spot.
        mean_photons_per_fluor : int
            Mean number of photons per fluorophore.

        Returns
        -------
        IntensityTable

        """

        # TODO nsofroniew: right now there is no jitter on x-y positions of the spots
        z = np.random.randint(0, num_z, size=n_spots)
        y = np.random.uniform(0, height - 1, size=n_spots)
        x = np.random.uniform(0, width - 1, size=n_spots)

        r = np.empty(n_spots)
        r.fill(np.nan)  # radius is a function of the point-spread gaussian size
        spot_attributes = SpotAttributes(
            pd.DataFrame(
                {Axes.ZPLANE.value: z,
                 Axes.Y.value: y,
                 Axes.X.value: x,
                 Features.SPOT_RADIUS: r}
            )
        )

        # empty data tensor
        data = np.zeros(shape=(n_spots, *codebook.shape[1:]))

        targets = np.random.choice(
            codebook.coords[Features.TARGET], size=n_spots, replace=True)
        expected_bright_locations = np.where(codebook.loc[targets])

        # create a binary matrix where "on" spots are 1
        data[expected_bright_locations] = 1

        # add physical properties of fluorescence
        data *= np.random.poisson(mean_photons_per_fluor, size=data.shape)
        data *= np.random.poisson(mean_fluor_per_spot, size=data.shape)

        # convert data to float for consistency with starfish
        data = preserve_float_range(data)
        assert 0 < data.max() <= 1

        intensities = cls.from_spot_data(
            data, spot_attributes, np.arange(data.shape[1]), np.arange(data.shape[2]))
        return intensities

    @classmethod
    def from_image_stack(
            cls,
            image_stack,
            crop_x: int=0, crop_y: int=0, crop_z: int=0
    ) -> "IntensityTable":
        """Generate an IntensityTable from all the pixels in the ImageStack

        Parameters
        ----------
        crop_x : int
            Number of pixels to crop from both top and bottom of x.
        crop_y : int
            Number of pixels to crop from both top and bottom of y.
        crop_z : int
            Number of pixels to crop from both top and bottom of z.
        image_stack : ImageStack
            ImageStack containing pixels to be treated as intensities.

        Returns
        -------
        IntensityTable :
            IntensityTable containing one intensity per pixel (across channels and rounds).

        """

        # verify the image is large enough to crop
        assert crop_z * 2 < image_stack.shape['z']
        assert crop_y * 2 < image_stack.shape['y']
        assert crop_x * 2 < image_stack.shape['x']

        zmin = image_stack.axis_labels(Axes.ZPLANE)[crop_z]
        ymin = crop_y
        xmin = crop_x
        zmax = image_stack.axis_labels(Axes.ZPLANE)[-crop_z - 1]
        ymax = image_stack.shape['y'] - crop_y
        xmax = image_stack.shape['x'] - crop_x
        cropped_stack = image_stack.sel({Axes.ZPLANE: (zmin, zmax),
                                         Axes.Y: (ymin, ymax),
                                         Axes.X: (xmin, xmax)})

        data = cropped_stack.xarray.transpose(
            Axes.ZPLANE.value,
            Axes.Y.value,
            Axes.X.value,
            Axes.ROUND.value,
            Axes.CH.value,
        )

        # (pixels, ch, round)
        intensity_data = data.values.reshape(-1, image_stack.num_rounds, image_stack.num_chs)

        # IntensityTable pixel coordinates
        z = image_stack.axis_labels(Axes.ZPLANE)
        y = np.arange(ymin, ymax)
        x = np.arange(xmin, xmax)

        feature_attribute_data = pd.DataFrame(
            data=np.array(list(product(z, y, x))),
            columns=['z', 'y', 'x']
        )
        feature_attribute_data[Features.SPOT_RADIUS] = np.full(
            feature_attribute_data.shape[0], fill_value=0.5
        )

        pixel_coordinates = SpotAttributes(feature_attribute_data)

        return IntensityTable.from_spot_data(
            intensity_data,
            pixel_coordinates,
            image_stack.axis_labels(Axes.ROUND),
            image_stack.axis_labels(Axes.CH),
        )

    @staticmethod
    def _process_overlaps(
        intensity_tables: List["IntensityTable"],
        overlap_strategy: OverlapStrategy
    ) -> List["IntensityTable"]:
        """
        Find the overlapping sections between IntensityTables and process them according
        to the given overlap strategy
        """
        overlap_pairs = find_overlaps_of_xarrays(intensity_tables)
        for indices in overlap_pairs:
            overlap_method = OVERLAP_STRATEGY_MAP[overlap_strategy]
            idx1, idx2 = indices
            # modify IntensityTables based on overlap strategy
            it1, it2 = overlap_method(intensity_tables[idx1], intensity_tables[idx2])
            # replace IntensityTables in list
            intensity_tables[idx1] = it1
            intensity_tables[idx2] = it2
        return intensity_tables

    @staticmethod
    def concatenate_intensity_tables(
        intensity_tables: List["IntensityTable"],
        overlap_strategy: Optional[OverlapStrategy] = None
    ) -> "IntensityTable":
        """
        Parameters
        ----------
        intensity_tables: List[IntensityTable]
            List of IntensityTables to be combined.
        overlap_strategy


        Returns
        -------

        """
        if overlap_strategy:
            intensity_tables = IntensityTable._process_overlaps(
                intensity_tables, overlap_strategy
            )
        return xr.concat(intensity_tables, dim=Features.AXIS)

    def to_features_dataframe(self) -> pd.DataFrame:
        """
        Generates a dataframe of the underlying features metadata.
        This is guaranteed to contain the features x, y, z, and radius.

        Returns
        -------
        pd.DataFrame
        """
        return pd.DataFrame(dict(self[Features.AXIS].coords))



from typing import MutableSequence, Tuple

import numpy as np


def fill_from_mask(
        mask: np.ndarray,
        offsets: Tuple[int, ...],
        fill_value: int,
        result_array: np.ndarray,
):
    """Take a binary mask and write `fill_value` to an array `result_array` where the binary mask
    has a True value.  The binary mask can have a different origin than the output array.  The
    relative offsets between the binary mask's origin an the output array's origin must be provided.

    Examples
    --------
    >>> import numpy as np
    >>> mask = np.array([True, True, False], dtype=np.bool)
    >>> mask
    array([ True,  True, False])
    >>> result_array = np.zeros(shape=(4,), dtype=np.uint32)
    >>> fill_from_mask(mask, (1,), 2, result_array)
    >>> result_array
    array([0, 2, 2, 0], dtype=uint32)
    """
    if mask.ndim != result_array.ndim:
        raise ValueError(
            f"mask ({mask.ndim}) should have the same number of dimensions as the result array "
            f"({result_array.ndim})")
    if mask.ndim != len(offsets):
        raise ValueError(
            f"mask ({mask.ndim}) should have the same number of dimensions as the number of "
            f"offsets ({len(offsets)})")

    _selector: MutableSequence[slice] = []
    for ix, (mask_size, result_size, axis_offset) in enumerate(
            zip(mask.shape, result_array.shape, offsets)):
        end_offset = axis_offset + mask_size
        if axis_offset < 0:
            raise ValueError(
                "{ix}th dimension has a negative offset ({axis_offset}), which is not permitted")
        if end_offset > result_size:
            raise ValueError(
                f"{ix}th dimension of mask does not fit within the result (ends at {end_offset}, "
                f"result size is {result_size}")

        _selector.append(slice(axis_offset, end_offset))
    selector: Tuple[slice, ...] = tuple(_selector)

    fill_value_array = result_array[selector]
    fill_value_array[mask] = fill_value
    result_array[selector] = fill_value_array



from functools import partial
from typing import Optional, Union

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Axes, Clip, Levels
from ._base import FilterAlgorithm


class LinearUnmixing(FilterAlgorithm):
    """
    LinearUnmixing enables the user to correct fluorescent bleed by subtracting fractions of the
    intensities of other channels from each channel in the ImageStack.

    Examples
    --------
    The following example provides a coefficient matrix that corrects for spectral
    mixing in a 3-channel experiment.
    Channel 0 contains a mixture of itself plus 50% of the intensity of
    channel 2. Channel 1 has no mixing with other channels. Channel 3
    consists of itself plus 10% of the intensity of both channels 0 and 1.

    >>> import numpy as np
    >>> coeff_mat = np.ndarray([
    ...     [1,    0,  -0.1]
    ...     [0,    1,  -0.1]
    ...     [-0.5, 0,  1   ]
    ... ])

    The end result of this unmixing will be that 50% of channel 2 will be subtracted from
    channel 0, channel 1 will not be changed, and 10% of channels 0 and 1 will be subtracted
    from channel 2.

    Parameters
    ----------
    coeff_mat : np.ndarray
        matrix of the linear unmixing coefficients. Should take the form: B = AX, where B are
        the unmixed values, A is coeff_mat and X are the observed values. coeff_mat has shape
        (n_ch, n_ch), and poses each channel (column) as a combination of other columns (rows).
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.
    """

    def __init__(
            self,
            coeff_mat: np.ndarray,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None
    ) -> None:
        self.coeff_mat = coeff_mat
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)

    _DEFAULT_TESTING_PARAMETERS = {"coeff_mat": np.array([[1, -0.25], [-0.25, 1]])}

    @staticmethod
    def _unmix(image: xr.DataArray, coeff_mat: np.ndarray) -> np.ndarray:
        """Perform linear unmixing of channels

        Parameters
        ----------
        image : np.ndarray
            image to be scaled

        coeff_mat : np.ndarray
            matrix of the linear unmixing coefficients. Should take the form:
            B = AX, where B are the unmixed values, A is coeff_mat and X are
            the observed values. coeff_mat has shape (n_ch, n_ch), and poses
            each channel (column) as a combination of other columns (rows).

        Returns
        -------
        np.ndarray :
          Numpy array of same shape as image

        """

        x = image.sizes[Axes.X.value]
        y = image.sizes[Axes.Y.value]
        c = image.sizes[Axes.CH.value]

        # broadcast each channel coefficient across x and y
        broadcast_coeff = np.tile(coeff_mat, reps=x * y).reshape(c, y, x, c)

        # multiply the image by each coefficient
        broadcast_image = image.values[..., None] * broadcast_coeff

        # collapse the unmixed result
        unmixed_image = np.sum(broadcast_image, axis=0).transpose([2, 0, 1])

        return unmixed_image

    def run(
            self,
            stack: ImageStack,
            in_place: bool=False,
            verbose: bool=False,
            n_processes: Optional[int]=None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = {Axes.ROUND, Axes.ZPLANE}
        unmix = partial(self._unmix, coeff_mat=self.coeff_mat)
        result = stack.apply(
            unmix,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
            level_method=self.level_method
        )
        return result



from ._base import FindSpotsAlgorithm
from .blob import BlobDetector
from .local_max_peak_finder import LocalMaxPeakFinder
from .trackpy_local_max_peak_finder import TrackpyLocalMaxPeakFinder

# autodoc's automodule directive only captures the modules explicitly listed in __all__.
all_filters = {
    filter_name: filter_cls
    for filter_name, filter_cls in locals().items()
    if isinstance(filter_cls, type) and issubclass(filter_cls, FindSpotsAlgorithm)
}
__all__ = list(all_filters.keys())



from collections import OrderedDict
from typing import Collection, List, Mapping, MutableSequence, Optional, Tuple, Union

import numpy as np
from slicedimage import Tile, TileSet

from starfish.core.imagestack.parser import TileCollectionData, TileData, TileKey
from starfish.core.types import ArrayLike, Axes, Coordinates, Number


class CropParameters:
    """Parameters for cropping an ImageStack at load time."""
    def __init__(
            self,
            *,
            permitted_rounds: Optional[Collection[int]]=None,
            permitted_chs: Optional[Collection[int]]=None,
            permitted_zplanes: Optional[Collection[int]]=None,
            x_slice: Optional[Union[int, slice]]=None,
            y_slice: Optional[Union[int, slice]]=None,
    ):
        """
        Parameters
        ----------
        permitted_rounds : Optional[Collection[int]]
            The rounds in the original dataset to load into the ImageStack.  If this is not set,
            then all rounds are loaded into the ImageStack.
        permitted_chs : Optional[Collection[int]]
            The channels in the original dataset to load into the ImageStack.  If this is not set,
            then all channels are loaded into the ImageStack.
        permitted_zplanes : Optional[Collection[int]]
            The z-layers in the original dataset to load into the ImageStack.  If this is not set,
            then all z-layers are loaded into the ImageStack.
        x_slice : Optional[Union[int, slice]]
            The x-range in the x-y tile that is loaded into the ImageStack.  If this is not set,
            then the entire x-y tile is loaded into the ImageStack.
        y_slice : Optional[Union[int, slice]]
            The y-range in the x-y tile that is loaded into the ImageStack.  If this is not set,
            then the entire x-y tile is loaded into the ImageStack.
        """
        self._permitted_rounds = set(permitted_rounds) if permitted_rounds else None
        self._permitted_chs = set(permitted_chs) if permitted_chs else None
        self._permitted_zplanes = set(permitted_zplanes) if permitted_zplanes else None
        self._x_slice = x_slice
        self._y_slice = y_slice

    def _add_permitted_axes(self, axis_type: Axes, permitted_axis: int) -> None:
        """
        Add a value to one of the permitted axes sets.
        """
        if axis_type == Axes.ROUND and self._permitted_rounds:
            self._permitted_rounds.add(permitted_axis)
        if axis_type == Axes.CH and self._permitted_chs:
            self._permitted_chs.add(permitted_axis)
        if axis_type == Axes.ZPLANE and self._permitted_zplanes:
            self._permitted_zplanes.add(permitted_axis)

    def filter_tilekeys(self, tilekeys: Collection[TileKey]) -> Collection[TileKey]:
        """
        Filters tilekeys for those that should be included in the resulting ImageStack.
        """
        results: MutableSequence[TileKey] = list()
        for tilekey in tilekeys:
            if self._permitted_rounds is not None and tilekey.round not in self._permitted_rounds:
                continue
            if self._permitted_chs is not None and tilekey.ch not in self._permitted_chs:
                continue
            if self._permitted_zplanes is not None and tilekey.z not in self._permitted_zplanes:
                continue

            results.append(tilekey)

        return results

    @staticmethod
    def _crop_axis(size: int, crop: Optional[Union[int, slice]]) -> Tuple[int, int]:
        """
        Given the size of along an axis, and an optional cropping, return the start index
        (inclusive) and end index (exclusive) of the crop.  If no crop is specified, then the
        original size (0, size) is returned.
        """
        # convert int crops to a slice operation.
        if isinstance(crop, int):
            if crop < 0 or crop >= size:
                raise IndexError("crop index out of range")
            return crop, crop + 1

        # convert start and stop to absolute values.
        start: int
        if crop is None or crop.start is None:
            start = 0
        elif crop.start is not None and crop.start < 0:
            start = max(0, size + crop.start)
        else:
            start = min(size, crop.start)

        stop: int
        if crop is None or crop.stop is None:
            stop = size
        elif crop.stop is not None and crop.stop < 0:
            stop = max(0, size + crop.stop)
        else:
            stop = min(size, crop.stop)

        return start, stop

    @staticmethod
    def parse_aligned_groups(tileset: TileSet,
                             rounds: Optional[Collection[int]] = None,
                             chs: Optional[Collection[int]] = None,
                             zplanes: Optional[Collection[int]] = None,
                             x: Optional[Union[int, slice]] = None,
                             y: Optional[Union[int, slice]] = None
                             ) -> List["CropParameters"]:

        """Takes a tileset and any optional selected axes lists compares the physical coordinates on each
         tile to create aligned coordinate groups (groups of tiles that have the same physical
         coordinates)

        Parameters
        ----------
        tileset: TileSet
            The TileSet to parse
        rounds: Optional[Collection[int]]
            The rounds in the tileset to include in the final aligned groups. If this is not set,
            then all rounds are included.
        chs: Optional[Collection[int]]
            The chs in the tileset to include in the final aligned groups. If this is not set,
            then all chs are included.
        zplanes: Optional[Collection[int]]
            The zplanes in the tileset to include in the final aligned groups. If this is not set,
            then all zplanes are included.
        x: Optional[Union[int, slice]]
            The x-range in the x-y tile to include in the final aligned groups.  If this is not set,
            then the entire x-y tile is included.
        y: Optional[Union[int, slice]]
            The y-range in the x-y tile to include in the final aligned groups.  If this is not set,
            then the entire x-y tile is included.

         Returns
         -------
         List["CropParameters"]
             A list of CropParameters. Each entry describes the r/ch/z values of tiles that are
             aligned (have matching coordinates) and are within the selected_axes if provided.
         """
        coord_groups: OrderedDict[tuple, CropParameters] = OrderedDict()
        for tile in tileset.tiles():
            if CropParameters.tile_in_selected_axes(tile, rounds, chs, zplanes):
                x_y_coords = (
                    tile.coordinates[Coordinates.X][0], tile.coordinates[Coordinates.X][1],
                    tile.coordinates[Coordinates.Y][0], tile.coordinates[Coordinates.Y][1]
                )
                # A tile with this (x, y) has already been seen, add tile's indices to
                # CropParameters
                if x_y_coords in coord_groups:
                    crop_params = coord_groups[x_y_coords]
                    crop_params._add_permitted_axes(Axes.CH, tile.indices[Axes.CH])
                    crop_params._add_permitted_axes(Axes.ROUND, tile.indices[Axes.ROUND])
                    if Axes.ZPLANE in tile.indices:
                        crop_params._add_permitted_axes(Axes.ZPLANE, tile.indices[Axes.ZPLANE])
                else:
                    coord_groups[x_y_coords] = CropParameters(
                        permitted_chs=[tile.indices[Axes.CH]],
                        permitted_rounds=[tile.indices[Axes.ROUND]],
                        permitted_zplanes=[tile.indices[Axes.ZPLANE]]
                        if Axes.ZPLANE in tile.indices else None,
                        x_slice=x,
                        y_slice=y)
        return list(coord_groups.values())

    @staticmethod
    def tile_in_selected_axes(tile: Tile,
                              rounds: Optional[Collection[int]] = None,
                              chs: Optional[Collection[int]] = None,
                              zplanes: Optional[Collection[int]] = None) -> bool:
        """
        Return True if a tile belongs in a list of selected axes.

        Parameters
        ----------
        tile:
            The tile in question
        rounds: Optional[Collection[int]]
            The allowed rounds.
        chs: Optional[Collection[int]]
            The allowed chs.
        zplanes: Optional[Collection[int]]
            The allowed zplanes.

        Returns
        -------
        Boolean
            True if tile belongs with selected axes, False if not.
        """
        if rounds and tile.indices[Axes.ROUND] not in rounds:
            return False
        if chs and tile.indices[Axes.CH] not in chs:
            return False
        if zplanes and tile.indices[Axes.ZPLANE] not in zplanes:
            return False
        return True

    def crop_shape(self, shape: Mapping[Axes, int]) -> Mapping[Axes, int]:
        """
        Given the shape of the original tile, return the shape of the cropped tile.
        """
        output_x_shape = CropParameters._crop_axis(shape[Axes.X], self._x_slice)
        output_y_shape = CropParameters._crop_axis(shape[Axes.Y], self._y_slice)
        width = output_x_shape[1] - output_x_shape[0]
        height = output_y_shape[1] - output_y_shape[0]

        return {Axes.Y: height, Axes.X: width}

    def crop_image(self, image: np.ndarray) -> np.ndarray:
        """
        Given the original image, return the cropped image.
        """
        output_x_shape = CropParameters._crop_axis(image.shape[1], self._x_slice)
        output_y_shape = CropParameters._crop_axis(image.shape[0], self._y_slice)

        return image[output_y_shape[0]:output_y_shape[1], output_x_shape[0]:output_x_shape[1]]

    def crop_coordinates(
            self, coordinates: Mapping[Coordinates, ArrayLike[Number]],
    ) -> Mapping[Coordinates, ArrayLike[Number]]:
        """
        Given a mapping of coordinate to coordinate values, return a mapping of the coordinate to
        cropped coordinate values.
        """
        output_x_shape = CropParameters._crop_axis(len(coordinates[Coordinates.X]), self._x_slice)
        output_y_shape = CropParameters._crop_axis(len(coordinates[Coordinates.Y]), self._y_slice)

        return_coords = {
            Coordinates.X: coordinates[Coordinates.X][output_x_shape[0]:output_x_shape[1]],
            Coordinates.Y: coordinates[Coordinates.Y][output_y_shape[0]:output_y_shape[1]],
        }

        if Coordinates.Z in coordinates:
            return_coords[Coordinates.Z] = coordinates[Coordinates.Z]

        return return_coords


class CroppedTileData(TileData):
    """Represent a cropped view of a TileData object."""
    def __init__(self, tile_data: TileData, cropping_parameters: CropParameters):
        self.backing_tile_data = tile_data
        self.cropping_parameters = cropping_parameters

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        return self.cropping_parameters.crop_shape(self.backing_tile_data.tile_shape)

    @property
    def numpy_array(self) -> np.ndarray:
        return self.cropping_parameters.crop_image(self.backing_tile_data.numpy_array)

    @property
    def coordinates(self) -> Mapping[Coordinates, ArrayLike[Number]]:
        return self.cropping_parameters.crop_coordinates(self.backing_tile_data.coordinates)

    @property
    def selector(self) -> Mapping[Axes, int]:
        return self.backing_tile_data.selector


class CroppedTileCollectionData(TileCollectionData):
    """Represent a cropped view of a TileCollectionData object."""
    def __init__(
            self,
            backing_tile_collection_data: TileCollectionData,
            crop_parameters: CropParameters,
    ) -> None:
        self.backing_tile_collection_data = backing_tile_collection_data
        self.crop_parameters = crop_parameters

    def __getitem__(self, tilekey: TileKey) -> dict:
        return self.backing_tile_collection_data[tilekey]

    def keys(self) -> Collection[TileKey]:
        return self.crop_parameters.filter_tilekeys(self.backing_tile_collection_data.keys())

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        return self.crop_parameters.crop_shape(self.backing_tile_collection_data.tile_shape)

    @property
    def extras(self) -> dict:
        return self.backing_tile_collection_data.extras

    def get_tile_by_key(self, tilekey: TileKey) -> TileData:
        return CroppedTileData(
            self.backing_tile_collection_data.get_tile_by_key(tilekey),
            self.crop_parameters,
        )

    def get_tile(self, r: int, ch: int, z: int) -> TileData:
        return CroppedTileData(
            self.backing_tile_collection_data.get_tile(r, ch, z),
            self.crop_parameters,
        )



from typing import Callable, Optional

from starfish.core.codebook.codebook import Codebook
from starfish.core.intensity_table.decoded_intensity_table import DecodedIntensityTable
from starfish.core.intensity_table.intensity_table_coordinates import \
    transfer_physical_coords_to_intensity_table
from starfish.core.spots.DecodeSpots.trace_builders import TRACE_BUILDERS
from starfish.core.types import SpotFindingResults, TraceBuildingStrategies
from ._base import DecodeSpotsAlgorithm


class PerRoundMaxChannel(DecodeSpotsAlgorithm):
    """
    Decode spots by selecting the max-valued channel in each sequencing round.

    Note that this assumes that the codebook contains only one "on" channel per sequencing round,
    a common pattern in experiments that assign one fluorophore to each DNA nucleotide and
    read DNA sequentially. It is also a characteristic of single-molecule FISH and RNAscope
    codebooks.

    Parameters
    ----------
    codebook : Codebook
        Contains codes to decode IntensityTable
    trace_building_strategy: TraceBuildingStrategies
        Defines the strategy for building spot traces to decode across rounds and chs of spot
        finding results.
    search_radius : Optional[int]
        Only applicable if trace_building_strategy is TraceBuildingStrategies.NEAREST_NEIGHBORS.
        Number of pixels over which to search for spots in other rounds and channels.
    anchor_round : Optional[int]
        Only applicable if trace_building_strategy is TraceBuildingStrategies.NEAREST_NEIGHBORS.
        The imaging round against which other rounds will be checked for spots in the same
        approximate pixel location.
    """

    def __init__(
            self,
            codebook: Codebook,
            trace_building_strategy: TraceBuildingStrategies = TraceBuildingStrategies.EXACT_MATCH,
            anchor_round: Optional[int]=1,
            search_radius: Optional[int]=3):
        self.codebook = codebook
        self.trace_builder: Callable = TRACE_BUILDERS[trace_building_strategy]
        self.anchor_round = anchor_round
        self.search_radius = search_radius

    def run(self, spots: SpotFindingResults, *args) -> DecodedIntensityTable:
        """Decode spots by selecting the max-valued channel in each sequencing round

        Parameters
        ----------
        spots: SpotFindingResults
            A Dict of tile indices and their corresponding measured spots

        Returns
        -------
        DecodedIntensityTable :
            IntensityTable decoded and appended with Features.TARGET and Features.QUALITY values.

        """
        intensities = self.trace_builder(spot_results=spots,
                                         anchor_round=self.anchor_round,
                                         search_radius=self.search_radius)
        transfer_physical_coords_to_intensity_table(intensity_table=intensities, spots=spots)
        return self.codebook.decode_per_round_max(intensities)



import json
from typing import (
    Any,
    Callable,
    Collection,
    Iterator,
    List,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Optional,
    Sequence,
    Set,
    Union
)

from semantic_version import Version
from slicedimage import Collection as collec
from slicedimage import TileSet
from slicedimage.io import Reader, resolve_path_or_url, resolve_url
from slicedimage.urlpath import pathjoin

from starfish.core.codebook.codebook import Codebook
from starfish.core.config import StarfishConfig
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.imagestack.parser.crop import CropParameters
from starfish.core.spacetx_format import validate_sptx
from .version import MAX_SUPPORTED_VERSION, MIN_SUPPORTED_VERSION

_SINGLETON = object()


class FieldOfView:
    """
    This encapsulates a field of view.  It contains the primary image and auxiliary images that are
    associated with the field of view.

    All images can be accessed using
    :py:func:`~starfish.experiment.experiment.FieldOfView.get_image`
    with the name of the image type and any which rounds/chs/zplanes you want to include
    (if none are provided all are included) If the resulting tileset has unaligned tiles
    (their x/y coordinates do not all match) we return a list on ImageStacks where each stack
    represents an aligned group of tiles. The primary image is accessed using the name
    :py:attr:`~starfish.experiment.experiment.FieldOFView.PRIMARY_IMAGES`.

    Notes
    -----
    Field of views obtain their primary image from a :py:class:`~slicedimage.TileSet`.
    They can obtain their auxiliary image dictionary from a dictionary of auxiliary image to
    :py:class:`~slicedimage.TileSet`.

    The decoding of :py:class:`~slicedimage.TileSet` to
    :py:class:`~starfish.imagestack.imagestack.ImageStack`
    does not happen until the image is accessed. ImageStacks can only be initialized with
    aligned tilesets, so if there are multiple you may need to iterate through the groups
    using :py:func:`~starfish.experiment.experiment.FieldOfView.get_images`
    and process each one individually.

    Be prepared to handle errors when images are accessed.

    Access a FieldOfView through Experiment.
    :py:func:`~starfish.experiment.experiment.Experiment.fov`.

    Attributes
    ----------
    name : str
        The name of the FOV.
    image_types : Set[str]
        A set of all the image types.
    """

    PRIMARY_IMAGES = 'primary'

    def __init__(
            self, name: str,
            image_tilesets: MutableMapping[str, TileSet]
    ) -> None:
        self._images: MutableMapping[str, TileSet] = dict()
        self._name = name
        self._images = image_tilesets

    def __repr__(self):
        images = '\n    '.join(
            f'{k}: {v}'
            for k, v in self._images.items()
            if k != FieldOfView.PRIMARY_IMAGES
        )
        return (
            f"<starfish.FieldOfView>\n"
            f"  Primary Image: {self._images[FieldOfView.PRIMARY_IMAGES]}\n"
            f"  Auxiliary Images:\n"
            f"    {images}"
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def image_types(self) -> Set[str]:
        return set(self._images.keys())

    def get_image(self, item: str,
                  aligned_group: Any = _SINGLETON,
                  rounds: Optional[Collection[int]] = None,
                  chs: Optional[Collection[int]] = None,
                  zplanes: Optional[Collection[int]] = None,
                  x: Optional[Union[int, slice]] = None,
                  y: Optional[Union[int, slice]] = None,
                  ) -> ImageStack:
        """
        Load into memory the first Imagestack representation of an aligned image group. If crop
        parameters provided, first crop the TileSet.

        Parameters
        ----------
        item: str
            The name of the tileset ex. 'primary' or 'nuclei'
        rounds : Optional[Collection[int]]
            The rounds in the original dataset to load into the ImageStack.  If this is not set,
            then all rounds are loaded into the ImageStack.
        chs : Optional[Collection[int]]
            The channels in the original dataset to load into the ImageStack.  If this is not set,
            then all channels are loaded into the ImageStack.
        zplanes : Optional[Collection[int]]
            The z-layers in the original dataset to load into the ImageStack.  If this is not set,
            then all z-layers are loaded into the ImageStack.
        x : Optional[Union[int, slice]]
            The x-range in the x-y tile that is loaded into the ImageStack.  If this is not set,
            then the entire x-y tile is loaded into the ImageStack.
        y : Optional[Union[int, slice]]
            The y-range in the x-y tile that is loaded into the ImageStack.  If this is not set,
            then the entire x-y tile is loaded into the ImageStack.

        Returns
        -------
        ImageStack
            The instantiated image stack

        """
        if aligned_group != _SINGLETON:
            raise RuntimeError("The parameter 'aligned_group` is no longer accepted. Please use"
                               "FieldOfView.get_images() and  provide sets of selected axes "
                               "instead")
        stack_iterator = self.get_images(item=item, rounds=rounds,
                                         chs=chs, zplanes=zplanes, x=x, y=y)
        return next(stack_iterator)

    def get_images(self, item: str,
                   rounds: Optional[Collection[int]] = None,
                   chs: Optional[Collection[int]] = None,
                   zplanes: Optional[Collection[int]] = None,
                   x: Optional[Union[int, slice]] = None,
                   y: Optional[Union[int, slice]] = None,
                   ) -> Iterator[ImageStack]:
        """
        Load into memory the an iterator of aligned Imagestacks for the given tileset and selected
        axes.

        Parameters
        ----------
        item: str
            The name of the tileset ex. 'primary' or 'nuclei'
        rounds : Optional[Collection[int]]
            The rounds in the original dataset to load into the ImageStack/s.  If this is not set,
            then all rounds are loaded into the ImageStack.
        chs : Optional[Collection[int]]
            The channels in the original dataset to load into the ImageStac/s.  If this is not set,
            then all channels are loaded into the ImageStack.
        zplanes : Optional[Collection[int]]
            The z-layers in the original dataset to load into the ImageStack/s.  If this is not set,
            then all z-layers are loaded into the ImageStack.
        x : Optional[Union[int, slice]]
            The x-range in the x-y tile that is loaded into the ImageStack/s.  If this is not set,
            then the entire x-y tile is loaded into the ImageStack.
        y : Optional[Union[int, slice]]
            The y-range in the x-y tile that is loaded into the ImageStack/s.  If this is not set,
            then the entire x-y tile is loaded into the ImageStack.

        Returns
        -------
        AlignedImageStackIterator
            The instantiated ImageStack or list of Imagestacks if the parameters given include
            multiple aligned groups.

        """
        # Parse the tileset into aligned groups, only include tiles from selected axes.
        aligned_groups = CropParameters.parse_aligned_groups(self._images[item],
                                                             rounds=rounds, chs=chs,
                                                             zplanes=zplanes,
                                                             x=x, y=y)
        aligned_stack_iterator = AlignedImageStackIterator(tileset=self._images[item],
                                                           aligned_groups=aligned_groups)
        return aligned_stack_iterator


class AlignedImageStackIterator(Iterator[ImageStack]):
    """Iterator class of AlignedImageStacks."""
    def __init__(self, tileset: TileSet, aligned_groups: List[CropParameters]):
        self.size = len(aligned_groups)
        self.aligned_groups = iter(aligned_groups)
        self.tileset = tileset

    def __len__(self) -> int:
        return self.size

    def __next__(self) -> ImageStack:
        aligned_group = next(self.aligned_groups)
        stack = ImageStack.from_tileset(self.tileset, aligned_group)
        return stack


class Experiment:
    """
    Encapsulates an experiment, with one or more fields of view and a
    :py:class:`~starfish.codebook.codebook.Codebook`. An individual
    :py:class:`~starfish.experiment.experiment.FieldOfView` can be retrieved using a
    key, i.e., experiment[fov_name].

    Attributes
    ----------
    codebook : Codebook
        The codebook associated with this experiment.
    extras : Dict
        Returns the extras dictionary associated with this experiment.
    """
    def __init__(
            self,
            fovs: Sequence[FieldOfView],
            codebook: Codebook,
            extras: dict,
            *,
            src_doc: dict=None,
    ) -> None:
        self._fovs = fovs
        self._codebook = codebook
        self._extras = extras
        self._src_doc = src_doc

    def __repr__(self):

        # truncate the list of fields of view if it is longer than print_n_fov
        print_n_fov = 4
        n_fields_of_view = list(self.items())[:print_n_fov]
        fields_of_view_str = "\n".join(
            f'{k}: {v}' for k, v in n_fields_of_view
        )

        # add an ellipsis if not all fields of view are being printed
        if len(self._fovs) > print_n_fov:
            fov_repr = f"{{\n{fields_of_view_str}\n  ...,\n}}"
        else:
            fov_repr = f"{{\n{fields_of_view_str}\n}}"

        # return the formatted string
        object_repr = f"<starfish.Experiment (FOVs={len(self._fovs)})>\n"
        return object_repr + fov_repr

    @classmethod
    def from_json(cls, json_url: str) -> "Experiment":
        """
        Construct an Experiment from an experiment.json file format specifier.
        Loads configuration from StarfishConfig.

        Parameters
        ----------
        json_url : str
            file path or web link to an experiment.json file

        Returns
        -------
        Experiment :
            Experiment object serving the requested experiment data

        """

        config = StarfishConfig()

        if config.strict:
            valid = validate_sptx.validate(json_url)
            if not valid:
                raise Exception("validation failed")

        backend, name, baseurl = resolve_path_or_url(json_url, config.slicedimage)
        with backend.read_contextmanager(name) as fh:
            experiment_document = json.load(fh)

        version = cls.verify_version(experiment_document['version'])

        _, codebook_name, codebook_baseurl = resolve_url(experiment_document['codebook'],
                                                         baseurl, config.slicedimage)
        codebook_absolute_url = pathjoin(codebook_baseurl, codebook_name)
        codebook = Codebook.open_json(codebook_absolute_url)

        extras = experiment_document['extras']

        fovs: MutableSequence[FieldOfView] = list()
        fov_tilesets: MutableMapping[str, TileSet]
        if version < Version("5.0.0"):
            primary_image: collec = Reader.parse_doc(experiment_document['primary_images'],
                                                     baseurl, config.slicedimage)
            auxiliary_images: MutableMapping[str, collec] = dict()
            for aux_image_type, aux_image_url in experiment_document['auxiliary_images'].items():
                auxiliary_images[aux_image_type] = Reader.parse_doc(
                    aux_image_url, baseurl, config.slicedimage)

            for fov_name, primary_tileset in primary_image.all_tilesets():
                fov_tilesets = dict()
                fov_tilesets[FieldOfView.PRIMARY_IMAGES] = primary_tileset
                for aux_image_type, aux_image_collection in auxiliary_images.items():
                    aux_image_tileset = aux_image_collection.find_tileset(fov_name)
                    if aux_image_tileset is not None:
                        fov_tilesets[aux_image_type] = aux_image_tileset

                fov = FieldOfView(fov_name, image_tilesets=fov_tilesets)
                fovs.append(fov)
        else:
            images: MutableMapping[str, collec] = dict()
            all_fov_names: MutableSet[str] = set()
            for image_type, image_url in experiment_document['images'].items():
                image = Reader.parse_doc(image_url, baseurl, config.slicedimage)
                images[image_type] = image
                for fov_name, _ in image.all_tilesets():
                    all_fov_names.add(fov_name)

            for fov_name in all_fov_names:
                fov_tilesets = dict()
                for image_type, image_collection in images.items():
                    image_tileset = image_collection.find_tileset(fov_name)
                    if image_tileset is not None:
                        fov_tilesets[image_type] = image_tileset

                fov = FieldOfView(fov_name, image_tilesets=fov_tilesets)
                fovs.append(fov)

        return Experiment(fovs, codebook, extras, src_doc=experiment_document)

    @classmethod
    def verify_version(cls, semantic_version_str: str) -> Version:
        """Verifies the compatibility of the current starfish version with
        the experiment version"""
        version = Version(semantic_version_str)
        if not (MIN_SUPPORTED_VERSION <= version <= MAX_SUPPORTED_VERSION):
            raise ValueError(
                f"version {version} not supported.  This version of the starfish library only "
                f"supports formats from {MIN_SUPPORTED_VERSION} to "
                f"{MAX_SUPPORTED_VERSION}")
        return version

    def fov(
            self,
            filter_fn: Callable[[FieldOfView], bool]=lambda _: True,
            key_fn: Callable[[FieldOfView], str]=lambda fov: fov.name,
    ) -> FieldOfView:
        """
        Given a callable filter_fn, apply it to all the FOVs in this experiment.

        Parameters
        ----------
        filter_fn : Callable
            Filter to apply to the list of FOVs
        key_fn : Callable
            The key that determines the order of filtered FOVs, default fov.name

        Returns
        -------
        FieldOfView
            The first FOV that fulfills the filter parameters.

        Raises
        ------
        LookupError :
             If no FOV matches

        """
        for fov in sorted(self._fovs, key=key_fn):
            if filter_fn(fov):
                return fov
        raise LookupError("Cannot find any FOV that the filter allows.")

    def fovs(
            self,
            filter_fn: Callable[[FieldOfView], bool]=lambda _: True,
            key_fn: Callable[[FieldOfView], str]=lambda fov: fov.name,
    ) -> Sequence[FieldOfView]:
        """
        Given a callable filter_fn, apply it to all the FOVs in this experiment.

        Parameters
        ----------
        filter_fn : Callable
            Filter to apply to the list of FOVs
        key_fn : Callable
            The key that determines the order of filtered FOVs, default fov.name

        Returns
        -------
        Sequence[FieldOfView]
            All fovs that pass the filter function.

        """
        results: MutableSequence[FieldOfView] = list()
        for fov in self._fovs:
            if not filter_fn(fov):
                continue

            results.append(fov)
        results = sorted(results, key=key_fn)
        return results

    def fovs_by_name(
        self,
        *names,
    ) -> Sequence[FieldOfView]:
        """
        Given a name or set of names, return all fovs that match.

        Parameters
        ----------
        names : str
            The fov names to search for.

        Returns
        -------
        Sequence[FieldOfView]
            All fovs that match the given names.
        """
        return self.fovs(filter_fn=lambda fov: fov.name in names)

    def __getitem__(self, item):
        fovs = self.fovs_by_name(item)
        if len(fovs) == 0:
            raise IndexError(f"No field of view with name \"{item}\"")
        return fovs[0]

    def keys(self):
        """Return all fov names in the experiment"""
        return (fov.name for fov in self.fovs())

    def values(self):
        """Return all FieldOfViews in the experiment"""
        return (fov for fov in self.fovs())

    def items(self):
        """Return all names and fovs in the experiment"""
        return ((fov.name, fov) for fov in self.fovs())

    @property
    def codebook(self) -> Codebook:
        return self._codebook

    @property
    def extras(self):
        return self._extras



import numpy as np

from starfish.core.expression_matrix.concatenate import concatenate
from starfish.core.expression_matrix.expression_matrix import ExpressionMatrix
from starfish.types import Features


def test_concatenate_two_expression_matrices():
    a_data = np.array(
        [[0, 1],
         [1, 0]]
    )
    b_data = np.array(
        [[0],
         [1]]
    )
    dims = [Features.CELLS, Features.GENES]
    a_coords = [(Features.CELLS, [0, 1]), (Features.GENES, ["x", "y"])]
    b_coords = [(Features.CELLS, [0, 1]), (Features.GENES, ["x"])]

    a = ExpressionMatrix(a_data, dims=dims, coords=a_coords)
    b = ExpressionMatrix(b_data, dims=dims, coords=b_coords)

    concatenated = concatenate([a, b])

    expected = np.array(
        [[0, 1],
         [1, 0],
         [0, np.nan],
         [1, np.nan]]
    )

    np.testing.assert_equal(concatenated.values, expected)



from pathlib import Path
from typing import (
    Mapping,
    Optional,
    Sequence,
    Union,
)
from warnings import warn

from starfish.core.morphology.binary_mask.binary_mask import BinaryMaskCollection, MaskData
from starfish.core.morphology.label_image import label_image as li
from starfish.core.types import ArrayLike, Axes, Coordinates, Number
from starfish.core.util.logging import Log


class SegmentationMaskCollection(BinaryMaskCollection):
    """Deprecated in favor of BinaryMaskCollection."""
    def __init__(
            self,
            pixel_ticks: Union[Mapping[Axes, ArrayLike[int]], Mapping[str, ArrayLike[int]]],
            physical_ticks: Union[Mapping[Coordinates, ArrayLike[Number]],
                                  Mapping[str, ArrayLike[Number]]],
            masks: Sequence[MaskData],
            log: Optional[Log],
    ):
        warn(
            f"{self.__class__.__name__} has been deprecated in favor of "
            f"{BinaryMaskCollection.__name__}",
            DeprecationWarning
        )
        super().__init__(pixel_ticks, physical_ticks, masks, log)

    @classmethod
    def from_label_image(cls, label_image: li.LabelImage) -> "BinaryMaskCollection":
        warn(
            f"{cls.__class__.__name__} has been deprecated in favor of "
            f"{BinaryMaskCollection.__name__}",
            DeprecationWarning
        )
        return BinaryMaskCollection.from_label_image(label_image)

    @classmethod
    def open_targz(cls, path: Union[str, Path]) -> "BinaryMaskCollection":
        warn(
            f"{cls.__class__.__name__} has been deprecated in favor of "
            f"{BinaryMaskCollection.__name__}",
            DeprecationWarning
        )
        return BinaryMaskCollection.open_targz(path)



import functools
import inspect
from abc import ABCMeta

from starfish.core.types._constants import STARFISH_EXTRAS_KEY


class AlgorithmBase(ABCMeta):
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        if not inspect.isabstract(cls):
            cls.run = AlgorithmBase.run_with_logging(cls.run)

    @staticmethod
    def run_with_logging(func):
        """
        This method extends each pipeline component.run() method to also log itself and
        runtime parameters to the IntensityTable and ImageStack objects. There are two
        scenarios for this method:
        """
        @functools.wraps(func)
        def helper(*args, **kwargs):
            result = func(*args, **kwargs)
            if result is not None:
                method_class_str = str(args[0].__class__)
                if 'ApplyTransform' in method_class_str or 'Filter' in method_class_str:
                    # Update the log on the resulting ImageStack
                    result.log.update_log(args[0])
                if 'FindSpots' in method_class_str:
                    # Update the log on the resulting SpotFindingResults
                    result.log.update_log(args[0])
                if 'DecodeSpots' in method_class_str:
                    # update log then transfer to DecodedIntensityTable
                    spot_results = kwargs['spots']
                    spot_results.log.update_log(args[0])
                    result.attrs[STARFISH_EXTRAS_KEY] = spot_results.log.encode()
                if 'DetectPixels' in method_class_str:
                    stack = args[1]
                    # update log with spot detection instance args[0]
                    stack.log.update_log(args[0])
                    # get resulting intensity table and set log
                    it = result[0]
                    it.attrs[STARFISH_EXTRAS_KEY] = stack.log.encode()
            return result
        return helper



import os
import warnings

from starfish.core.util.config import Config, NestedDict


def special_prefix(key):
    """
    All environment variables starting with these prefixes require
    special handling
    """
    for x in ("STARFISH_", "SLICEDIMAGE_"):
        if key.startswith(x):
            return True
    return False


class environ(object):
    """Overrides environment variables (prefixed with ``STARFISH_``)
    for the duration of the call.

    Examples
    --------
    Turn on validation:

        >>> from starfish.config import environ
        >>> from starfish import Experiment
        >>> with environ(VALIDATION_STRICT="true"):
        >>>     Experiment.from_json(URL)

    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __enter__(self):
        self.orig = dict()
        for k, newval in self.kwargs.items():
            if not special_prefix(k):
                k = "STARFISH_%s" % k
            old = os.environ.get(k, None)
            try:
                os.environ[k] = newval
                self.orig[k] = old  # Only store if successful
            except TypeError:
                raise
        return self

    def __exit__(self, *args):
        for k, oldval in self.orig.items():
            if oldval is None:
                del os.environ[k]
            else:
                os.environ[k] = oldval


class StarfishConfig(object):
    """
    Application specific configuration settings which can be loaded throughout
    the starfish codebase.

    Attributes
    ----------
    slicedimage : dictionary
        Subdictionary that can be passed to slicedimage.io methods.
    strict : bool
        Whether or not loaded json should be validated.
    verbose : bool
        Controls output like from tqdm

    Examples
    --------
    Check strict property

        >>> from starfish.config import StarfishConfig
        >>> config = StarfishConfig()
        >>> if config.strict:
        >>>     validate(json)

    Default starfish configuration equivalent:

        >>> {
        >>>     "slicedimage": {
        >>>         "caching": {
        >>>             "debug": false,
        >>>             "directory": "~/.starfish/cache",
        >>>             "size_limit": 5e9
        >>>         },
        >>>     },
        >>>     "validation": {
        >>>         "strict": false
        >>>     },
        >>>     "verbose": true
        >>> }

    Example of a ~/.starfish.config file to disable caching:

        >>> {
        >>>     "slicedimage": {
        >>>         "caching": {
        >>>             "size_limit": 0
        >>>         }
        >>>     }
        >>> }

    """

    def __init__(self) -> None:
        """
        Loads the configuration specified by the STARFISH_CONFIG environment variable.

        Parameters
        ----------
        STARISH_CONFIG :
            This parameter is read from the environment to permit setting configuration
            values either directly or via a file. Keys read include:

             - ["slicedimage"]["caching"]["directory"]   (default: ~/.starfish/cache)
             - ["slicedimage"]["caching"]["size_limit"]  (default: None; 0 disables caching)
             - ["validation"]["strict"]                  (default: False)
             - ["verbose"]                               (default: True)

            Note: all keys can also be set by and environment variable constructed from the
            key parts and prefixed with STARFISH, e.g. STARFISH_VALIDATION_STRICT.
        """
        config = os.environ.get("STARFISH_CONFIG", "@~/.starfish/config")
        self._config_obj = Config(config)
        self._env_keys = [
            x for x in os.environ.keys()
            if special_prefix(x) and x != "STARFISH_CONFIG"]

        # If no directory is set, then force the default
        self._slicedimage = self._config_obj.lookup(("slicedimage",), NestedDict(), remove=True)
        if not self._slicedimage["caching"]["directory"]:
            self._slicedimage["caching"]["directory"] = "~/.starfish/cache"
        self._slicedimage_update(('caching', 'directory'))
        self._slicedimage_update(('caching', 'size_limit'), int)

        self._strict = self._config_obj.lookup(
            ("validation", "strict"), self.flag("STARFISH_VALIDATION_STRICT", "false"), remove=True)

        self._verbose = self._config_obj.lookup(
            ("verbose",), self.flag("STARFISH_VERBOSE", "true"), remove=True)

        if self._config_obj.data:
            warnings.warn(f"unknown configuration: {self._config_obj.data}")
        if self._env_keys:
            warnings.warn(f"unknown environment variables: {self._env_keys}")

    def _slicedimage_update(self, lookup, parse=lambda x: x):
        """ accept STARFISH_SLICEDIMAGE_ or SLICEDIMAGE_ prefixes"""

        value = None

        name1 = "SLICEDIMAGE_" + "_".join([x.upper() for x in lookup])
        if name1 in os.environ:
            self._env_keys.remove(name1)
            value = parse(os.environ[name1])

        name2 = "STARFISH_" + name1
        if name2 in os.environ:
            if value:
                warnings.warn(f"duplicate variable: (STARFISH_){name1}")
            self._env_keys.remove(name2)
            value = parse(os.environ[name2])

        if value is None:
            return  # Nothing found

        v = self._slicedimage
        for k in lookup[:-1]:
            v = v[k]
        v[lookup[-1]] = value

    def flag(self, name, default_value=""):

        if name in os.environ:
            value = os.environ[name]
            self._env_keys.remove(name)
        else:
            value = default_value

        if isinstance(value, str):
            value = value.lower()
            return value in ("true", "1", "yes", "y", "on", "active", "enabled")

    @property
    def slicedimage(self):
        return dict(self._slicedimage)

    @property
    def strict(self):
        return self._strict

    @property
    def verbose(self):
        return self._verbose



from typing import MutableSequence, Optional

from starfish.core.morphology.binary_mask import BinaryMaskCollection, MaskData
from ._base import FilterAlgorithm


class AreaFilter(FilterAlgorithm):
    """
    Create a BinaryMaskCollection using only the masks that meet the minimum and maximum area
    criteria.

    Parameters
    ----------
    min_area : Optional[int]
        Only masks that have area larger or equal to this will be included in the output mask
        collection.
    max_area : Optional[int]
        Only masks that have area smaller or equal to this will be included in the output mask
        collection.
    """

    def __init__(self, min_area: Optional[int], max_area: Optional[int]):
        self._min_area = min_area
        self._max_area = max_area

    def run(
            self,
            binary_mask_collection: BinaryMaskCollection,
            *args, **kwargs) -> BinaryMaskCollection:
        matching_mask_data: MutableSequence[MaskData] = list()
        for ix in range(len(binary_mask_collection)):
            props = binary_mask_collection.mask_regionprops(ix)
            if self._min_area is not None and props.area < self._min_area:
                continue
            if self._max_area is not None and props.area > self._max_area:
                continue

            matching_mask_data.append(binary_mask_collection._masks[ix])

        return BinaryMaskCollection(
            binary_mask_collection._pixel_ticks,
            binary_mask_collection._physical_ticks,
            matching_mask_data,
            binary_mask_collection._log,
        )



from abc import abstractmethod

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.morphology.binary_mask import BinaryMaskCollection
from starfish.core.pipeline.algorithmbase import AlgorithmBase


class BinarizeAlgorithm(metaclass=AlgorithmBase):

    @abstractmethod
    def run(self, image: ImageStack, *args, **kwargs) -> BinaryMaskCollection:
        """Performs binarization on the stack provided."""
        raise NotImplementedError()



class DeprecatedAPIError(Exception):
    """
    Raised when using an API that has been deprecated.
    """
    pass


class DataFormatWarning(Warning):
    """
    Warnings given by starfish when the data is not formatted as expected, though not fatally.
    """
    pass



from typing import Sequence, Tuple, TypeVar, Union

import numpy as np
import xarray as xr

from ._constants import (
    Axes,
    Clip,
    Coordinates,
    CORE_DEPENDENCIES,
    Features,
    Levels,
    LOG,
    OverlapStrategy,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes,
    STARFISH_EXTRAS_KEY,
    TraceBuildingStrategies,
    TransformType
)
from ._decoded_spots import DecodedSpots
from ._functionsource import FunctionSource, FunctionSourceBundle
from ._spot_attributes import SpotAttributes
from ._spot_finding_results import PerImageSliceSpotResults, SpotFindingResults

Number = Union[int, float]
CoordinateValue = Union[Number, Tuple[Number, Number]]

NPArrayLike = TypeVar("NPArrayLike", np.ndarray, xr.DataArray)

ArrayLikeTypes = TypeVar("ArrayLikeTypes", int, Number)
ArrayLike = Union[xr.DataArray, np.ndarray, Sequence[ArrayLikeTypes]]
"""ArrayLike is a parameterizable custom type that includes np.ndarrays, xr.DataArrays, and
Sequences of typed values.  Once the scipy stack supports typed arrays
(https://github.com/numpy/numpy/issues/7370), we can extend that to the array types."""



import numpy as np

from starfish import ImageStack
from starfish.core.image.Filter.linear_unmixing import LinearUnmixing
from starfish.core.types import Clip

def setup_linear_unmixing_test():
    """
        Create the image stack, coeff matrix, and reference result
        for the linear unmixing test

    """
    # Create image
    r, c, z, y, x = 1, 3, 6, 5, 4
    im = np.ones((r, c, z, y, x), dtype=np.float32)

    # Create a single pixel with zero intensity
    im[0, 0, 0, 0, 0] = 0
    stack = ImageStack.from_numpy(im)

    # Create coefficients matrix
    coeff_mat = np.array(
        [[ 1.00, -0.25,  0.00],  # noqa
         [-0.25,  1.00, -0.25],  # noqa
         [-0.10,  0.00,  1.00]]  # noqa
    )

    # Create reference result
    ref_result = im.copy()
    ref_result[:, 0, ...] = 0.65 * np.ones((z, y, x))
    ref_result[:, 1, ...] = 0.75 * np.ones((z, y, x))
    ref_result[:, 2, ...] = 0.75 * np.ones((z, y, x))

    # Account for the zero-value pixel
    ref_result[0, 0, 0, 0, 0] = 0
    ref_result[0, 1, 0, 0, 0] = 1

    return stack, coeff_mat, ref_result

def test_linear_unmixing():
    """ Test the linear unmixing filter """

    stack, coeff_mat, ref_result = setup_linear_unmixing_test()

    filter_unmix = LinearUnmixing(coeff_mat=coeff_mat, clip_method=Clip.CLIP)
    stack2 = filter_unmix.run(stack, in_place=False, verbose=False)

    assert np.allclose(ref_result, np.asarray(stack2.xarray))



from collections import OrderedDict

import pytest
from slicedimage import ImageFormat

from starfish.core.types import Axes, PhysicalCoordinateTypes
from .factories import imagestack_with_coords_factory
from .imagestack_test_utils import verify_physical_coordinates
from ..imagestack import ImageStack
from ..physical_coordinates import _get_physical_coordinates_of_z_plane

X_COORDS = 1, 2
Y_COORDS = 4, 6
Z_COORDS = 1, 3


@pytest.mark.parametrize("format,count", (
    (ImageFormat.TIFF, 6),
    (ImageFormat.NUMPY, 6),
))
def test_imagestack_export(tmpdir, format, count, recwarn):
    """
    Save a synthetic stack to files and check the results
    """
    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 1), (Axes.Y, 50), (Axes.X, 40)])

    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, X_COORDS[0]),
                                   (PhysicalCoordinateTypes.X_MAX, X_COORDS[1]),
                                   (PhysicalCoordinateTypes.Y_MIN, Y_COORDS[0]),
                                   (PhysicalCoordinateTypes.Y_MAX, Y_COORDS[1]),
                                   (PhysicalCoordinateTypes.Z_MIN, Z_COORDS[0]),
                                   (PhysicalCoordinateTypes.Z_MAX, Z_COORDS[1])])

    stack = imagestack_with_coords_factory(stack_shape, physical_coords)

    stack_json = tmpdir / "output.json"
    stack.export(str(stack_json), tile_format=format)
    files = list([x for x in tmpdir.listdir() if str(x).endswith(format.file_ext)])
    loaded_stack = ImageStack.from_path_or_url(str(stack_json))
    verify_physical_coordinates(
        loaded_stack,
        X_COORDS,
        Y_COORDS,
        _get_physical_coordinates_of_z_plane(Z_COORDS),
    )
    assert count == len(files)
    with open(files[0], "rb") as fh:
        format.reader_func(fh)



import numpy as np

from starfish import IntensityTable
from starfish.core.types import Coordinates, Features
from starfish.core.types._constants import OverlapStrategy
from .factories import create_intensity_table_with_coords
from ..overlap import (
    Area,
    find_overlaps_of_xarrays,
    remove_area_of_xarray,
    sel_area_of_xarray
)


def test_find_area_intersection():
    """
    Create various Area objects and verify their intersection are calculated correctly
    """
    area1 = Area(min_x=0, max_x=2, min_y=0, max_y=2)
    area2 = Area(min_x=1, max_x=2, min_y=1, max_y=3)
    intersection = Area.find_intersection(area1, area2)
    # intersection should be area with bottom point (1,1) and top point (2,2)
    assert intersection == Area(min_x=1, max_x=2, min_y=1, max_y=2)

    area2 = Area(min_x=3, max_x=5, min_y=3, max_y=5)
    intersection = Area.find_intersection(area1, area2)
    # no intersection
    assert intersection is None

    area2 = Area(min_x=0, max_x=5, min_y=3, max_y=5)
    intersection = Area.find_intersection(area1, area2)
    # area 2 right above area one
    assert intersection is None

    # try negatives
    area1 = Area(min_x=-1, max_x=1, min_y=0, max_y=2)
    area2 = Area(min_x=0, max_x=2, min_y=0, max_y=2)
    intersection = Area.find_intersection(area1, area2)
    assert intersection == Area(min_x=0, max_x=1, min_y=0, max_y=2)

    area2 = Area(min_x=-3, max_x=-2, min_y=0, max_y=2)
    intersection = Area.find_intersection(area1, area2)
    assert intersection is None


def test_find_overlaps_of_xarrays():
    """
    Create a list of overlapping IntensityTables and verify we identify the correct
    overlapping sections
    """
    # Create some overlapping intensity tables
    it0 = create_intensity_table_with_coords(Area(min_x=0, max_x=1,
                                                  min_y=0, max_y=1))
    it1 = create_intensity_table_with_coords(Area(min_x=.5, max_x=2,
                                                  min_y=.5, max_y=1.5))
    it2 = create_intensity_table_with_coords(Area(min_x=1.5, max_x=2.5,
                                                  min_y=0, max_y=1))
    it3 = create_intensity_table_with_coords(Area(min_x=0, max_x=1,
                                                  min_y=1, max_y=2))
    overlaps = find_overlaps_of_xarrays([it0, it1, it2, it3])
    # should have 4 total overlaps
    assert len(overlaps) == 4
    # overlap 1 between it0 and it1:
    assert (0, 1) in overlaps
    # overlap 1 between it0 and it1:
    assert (1, 2) in overlaps
    # overlap 3 between it1 and it3
    assert (1, 3) in overlaps
    # overlap 4 between it0 and it3
    assert (0, 3) in overlaps


def test_remove_area_of_xarray():
    """
    Tests removing a section of an IntensityTable defined by its physical area
    """
    it = create_intensity_table_with_coords(Area(min_x=0, max_x=2,
                                                 min_y=0, max_y=2), n_spots=10)

    area = Area(min_x=1, max_x=2, min_y=1, max_y=3)
    # grab some random coord values in this range
    removed_x = it.where(it.xc > 1, drop=True)[Coordinates.X.value].data[0]
    removed_y = it.where(it.yc > 1, drop=True)[Coordinates.X.value].data[3]

    it = remove_area_of_xarray(it, area)
    # assert coords from removed section are no longer in it
    assert not np.any(np.isclose(it[Coordinates.X.value], removed_x))
    assert not np.any(np.isclose(it[Coordinates.Y.value], removed_y))


def test_sel_area_of_xarray():
    """
    Tests selecting a section of an IntensityTable defined by its physical area
    """
    it = create_intensity_table_with_coords(Area(min_x=0, max_x=2, min_y=0, max_y=2), n_spots=10)

    area = Area(min_x=1, max_x=2, min_y=1, max_y=3)
    it = sel_area_of_xarray(it, area)

    # Assert new min/max values
    assert min(it[Coordinates.X.value]).data >= 1
    assert max(it[Coordinates.X.value]).data <= 2
    assert min(it[Coordinates.Y.value]).data >= 1
    assert max(it[Coordinates.X.value]).data <= 2


def test_take_max():
    """
    Create two overlapping IntensityTables with differing number of spots and verify that
    by concatenating them with the TAKE_MAX strategy we only include spots in the overlapping
    section from the IntensityTable that had the most.
    """
    it1 = create_intensity_table_with_coords(Area(min_x=0, max_x=2,
                                                  min_y=0, max_y=2), n_spots=10)
    it2 = create_intensity_table_with_coords(Area(min_x=1, max_x=2,
                                                  min_y=1, max_y=3), n_spots=20)

    concatenated = IntensityTable.concatenate_intensity_tables(
        [it1, it2], overlap_strategy=OverlapStrategy.TAKE_MAX)

    # The overlap section hits half of the spots from each intensity table, 5 from it1
    # and 10 from i21. It2 wins and the resulting concatenated table should have all the
    # spots from it2 (20) and 6 (one on the border) from it1 (6) for a total of 26 spots
    assert concatenated.sizes[Features.AXIS] == 26



from abc import abstractmethod

import numpy as np

from starfish.core.intensity_table.decoded_intensity_table import DecodedIntensityTable
from starfish.core.pipeline.algorithmbase import AlgorithmBase


class AssignTargetsAlgorithm(metaclass=AlgorithmBase):
    """
    AssignTargets assigns cell IDs to detected spots using an IntensityTable and
    SegmentationMaskCollection.
    """

    @abstractmethod
    def run(
            self,
            label_image: np.ndarray,
            decoded_intensity_table: DecodedIntensityTable,
            verbose: bool=False,
            in_place: bool=False,
    ) -> DecodedIntensityTable:
        """Performs target (e.g. gene) assignment given the spots and the regions."""
        raise NotImplementedError()



from abc import abstractmethod

from starfish.core.intensity_table.decoded_intensity_table import DecodedIntensityTable
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.types import SpotFindingResults


class DecodeSpotsAlgorithm(metaclass=AlgorithmBase):
    """Performs decoding on the spots found, using the codebook specified."""

    @abstractmethod
    def run(self, spot_attributes: SpotFindingResults, *args) -> DecodedIntensityTable:
        raise NotImplementedError()



import numpy as np
import pytest

from starfish import BinaryMaskCollection, display, LabelImage
from starfish.core.test.factories import SyntheticData
from starfish.types import Coordinates


sd = SyntheticData(
    n_ch=2,
    n_round=3,
    n_spots=1,
    n_codes=4,
    n_photons_background=0,
    background_electrons=0,
    camera_detection_efficiency=1.0,
    gray_level=1,
    ad_conversion_bits=16,
    point_spread_function=(2, 2, 2),
)

stack = sd.spots()
spots = sd.intensities()
label_image = LabelImage.from_label_array_and_ticks(
    np.random.randint(0, 4, size=(128, 128), dtype=np.uint8),
    None,
    {Coordinates.Y: np.arange(128), Coordinates.X: np.arange(128)},
    None,
)
masks = BinaryMaskCollection.from_label_image(label_image)


@pytest.mark.napari
@pytest.mark.parametrize('masks', [masks, None], ids=['masks', '     '])
@pytest.mark.parametrize('spots', [spots, None], ids=['spots', '     '])
@pytest.mark.parametrize('stack', [stack, None], ids=['stack', '     '])
def test_display(stack, spots, masks):
    import napari
    from qtpy.QtCore import QTimer
    from qtpy.QtWidgets import QApplication

    def run():
        app = QApplication.instance() or QApplication([])
        viewer = napari.Viewer()
        timer = QTimer()
        timer.setInterval(1000)
        timer.timeout.connect(app.quit)
        timer.start()
        display(stack, spots, masks, viewer=viewer)
        app.exec_()

    if stack is None and spots is None and masks is None:
        with pytest.raises(TypeError):
            run()
    else:
        run()



from functools import partial
from typing import Optional

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from starfish.types import Axes
from ._base import FilterAlgorithm


class CallBases(FilterAlgorithm):
    """
    The CallBases filter determines the nucleotide present in each pixel of each
    (round, channel). The pixel values in the resulting image are the base quality
    score. The base quality score is calculated as the intensity of the pixel
    divided by the L2 norm of the all channels for that pixel. The base
    call score as a range of (0, 1) with a value of 0  and 1 being no call and a
    perfect call, respectively.

    Parameters
    ----------
    intensity_threshold : float
        Minimum intensity a pixel must have to be called a base.
        Set to zero for no thresholding.
    quality_threshold : float
        Minimum base quality score a pixel must have to be called a base.
        Set to zero for no thresholding.
    """

    def __init__(
        self, intensity_threshold: float = 0, quality_threshold: float = 0
    ) -> None:

        self.intensity_threshold = intensity_threshold
        self.quality_threshold = quality_threshold

    _DEFAULT_TESTING_PARAMETERS = {"intensity_threshold": 0, "quality_threshold": 0}

    def _vector_norm(self, x: xr.DataArray, dim: Axes, ord=None):
        """
        Calculates the vector norm across a given dimension of an xarray.

        Parameters
        ----------
        x : xarray.DataArray
            xarray to calcuate the norm.
        dim : Axes
            The dimension of the xarray to perform the norm across
        ord : Optional[str]
            Order of the norm to be applied. Leave none for L2 norm.

            See numpy docs for details:
            https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.norm.html

        Returns
        -------
        result : xarray.DataArray
            The resulting xarray of norms

        """
        result = xr.apply_ufunc(
            np.linalg.norm, x, input_core_dims=[[dim.value]],
            kwargs={'ord': ord, 'axis': -1}
        )

        return result

    def _call_bases(
        self, image: xr.DataArray, intensity_threshold: float,
        quality_threshold: float
    ) -> xr.DataArray:
        """
        Determines the nucleotide present in each pixel of each
        (round, channel). The pixel values in the resulting image are the base quality
        score. The base quality score is calculated as the intensity of the pixel
        divided by the L2 norm of the all channels for that pixel. The base
        call score as a range of (0, 1) with a value of 0  and 1 being no call and a
        perfect call, respectively.

        Parameters
        ----------
        image : xr.DataArray
            Image for base calling.
            Should have the following dims: Axes.CH, Axes.X, Axes.Y
        intensity_threshold : float
            Minimum intensity a pixel must have to be called a base.
            Set to zero for no thresholding.
        quality_threshold : float
            Minimum intensity a pixel must have to be called a base.
            Set to zero for no thresholding.

        """
        # Get the maximum value for each round/z
        max_chan = image.argmax(dim=Axes.CH.value)
        max_values = image.max(dim=Axes.CH.value)

        # Get the norms for each pixel
        norms = self._vector_norm(x=image, dim=Axes.CH)

        # Calculate the base qualities
        base_qualities = max_values / norms

        # Filter the base call qualities
        base_qualities_filtered = xr.where(
            base_qualities < quality_threshold, 0, base_qualities
        )

        # Threshold the intensity values
        base_qualities_filtered = xr.where(
            max_values < intensity_threshold, 0, base_qualities_filtered
        )

        # Put the base calls in place
        base_calls = xr.full_like(other=image, fill_value=0)
        base_calls[max_chan] = base_qualities_filtered

        return base_calls

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack. Otherwise return the
            original stack.

        """

        group_by = {Axes.ROUND, Axes.ZPLANE}
        unmix = partial(
            self._call_bases, intensity_threshold=self.intensity_threshold,
            quality_threshold=self.quality_threshold
        )
        result = stack.apply(
            unmix,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes
        )
        return result



from typing import Optional, Tuple

import numpy as np
import pandas as pd

from starfish.core.expression_matrix.expression_matrix import ExpressionMatrix
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import (
    Axes,
    Coordinates,
    DecodedSpots,
    Features,
)


class DecodedIntensityTable(IntensityTable):
    """
    DecodedIntensityTable is a container for spot or pixel features extracted from image data
    that have been decoded. It is the primary output from starfish :py:class:`Decode` methods.

    An IntensityTable records the numeric intensity of a set of features in each
    :code:`(round, channel)` tile in which the feature is identified.
    The :py:class:`IntensityTable` has shape
    :code:`(n_feature, n_channel, n_round)`.

    Some :py:class:`SpotFinder` methods identify a position and search for Gaussian blobs in a
    small radius, only recording intensities if they are found in a given tile. Other
    :py:class:SpotFinder: approaches find blobs in a max-projection and measure them everywhere.
    As a result, some IntensityTables will be dense, and others will contain :code:`np.nan`
    entries where no feature was detected.

    Examples
    --------
    Create an IntensityTable using the ``synthetic_intensities`` method::

        >>> from starfish.core.test.factories import SyntheticData
        >>> sd = SyntheticData(n_ch=3, n_round=4, n_codes=2)
        >>> codes = sd.codebook()
        >>> sd.intensities(codebook=codes)
        <xarray.IntensityTable (features: 2, c: 3, h: 4)>
        array([[[    0.,     0.,     0.,     0.],
                [    0.,     0.,  8022., 12412.],
                [11160.,  9546.,     0.,     0.]],

               [[    0.,     0.,     0.,     0.],
                [    0.,     0., 10506., 10830.],
                [11172., 12331.,     0.,     0.]]])
        Coordinates:
        * features   (features) MultiIndex
        - z          (features) int64 7 3
        - y          (features) int64 14 32
        - x          (features) int64 32 15
        - r          (features) float64 nan nan
        * c          (c) int64 0 1 2
        * h          (h) int64 0 1 2 3
         target     (features) object 08b1a822-a1b4-4e06-81ea-8a4bd2b004a9 ...

        """

    __slots__ = ()

    @classmethod
    def from_intensity_table(
            cls,
            intensities: IntensityTable,
            targets: Tuple[str, np.ndarray],
            distances: Optional[Tuple[str, np.ndarray]] = None,
            passes_threshold: Optional[Tuple[str, np.ndarray]] = None):
        """
        Assign target values to intensities.

        Parameters
        ----------
        intensities : IntensityTable
            intensity_table to assign target values to
        targets : Tuple[str, np.ndarray]
            Target values to assign
        distances : Optional[Tuple[str, np.ndarray]]
            Corresponding array of distances from target for each feature
        passes_threshold : Optional[Tuple[str, np.ndarray]]
            Corresponding array of boolean values indicating if each itensity passed
            given thresholds.

        Returns
        -------
        DecodedIntensityTable
        """

        intensities = cls(intensities)
        intensities[Features.TARGET] = targets
        if distances:
            intensities[Features.DISTANCE] = distances
        if passes_threshold:
            intensities[Features.PASSES_THRESHOLDS] = passes_threshold
        return intensities

    def to_decoded_dataframe(self) -> DecodedSpots:
        """
        Generates a dataframe containing decoded spot information. Guaranteed to contain physical
        spot coordinates (z, y, x) and gene target. Does not contain pixel coordinates.
        """
        df = self.to_features_dataframe()
        pixel_coordinates = pd.Index([Axes.X, Axes.Y, Axes.ZPLANE])
        df = df.drop(pixel_coordinates.intersection(df.columns), axis=1).drop(Features.AXIS, axis=1)
        return DecodedSpots(df)

    def to_mermaid(self, filename: str) -> pd.DataFrame:
        """
        Writes a .csv.gz file in columnar format that is readable by MERMAID visualization
        software.

        To run MERMAID, follow the installation instructions for that repository and simply
        replace the data.csv.gz file with the output of this function.

        Parameters
        ----------
        filename : str
            Name for compressed-gzipped MERMAID data file. Should end in '.csv.gz'.

        Notes
        ------
        See also https://github.com/JEFworks/MERmaid

        """
        # construct the MERMAID dataframe. As MERMAID adds support for non-categorical variables,
        # additional columns can be added here
        df = self.to_features_dataframe()
        column_order = [
            Axes.X,
            Axes.Y,
            Features.TARGET,
            Features.TARGET,  # added twice to support simultaneous coding
        ]
        mermaid_data = df[column_order]

        # write to disk
        mermaid_data.to_csv(filename, compression='gzip', index=False)

    def to_expression_matrix(self) -> ExpressionMatrix:
        """
        Generates a cell x gene count matrix where each cell is annotated with spatial metadata.

        Requires that spots in the IntensityTable have been assigned to cells.

        Returns
        -------
        ExpressionMatrix :
            cell x gene expression table
        """
        if Features.CELL_ID not in self.coords:
            raise KeyError(
                "IntensityTable must have 'cell_id' assignments for each cell before this function "
                "can be called. See starfish.spots.AssignTargets.Label.")
        grouped = self.to_features_dataframe().groupby([Features.CELL_ID, Features.TARGET])
        counts = grouped.count().iloc[:, 0].unstack().fillna(0)
        if self.has_physical_coords:
            grouped = self.to_features_dataframe().groupby([Features.CELL_ID])[[
                Axes.X.value, Axes.Y.value, Axes.ZPLANE.value, Coordinates.X.value,
                Coordinates.Y.value, Coordinates.Z.value]]
        else:
            grouped = self.to_features_dataframe().groupby([Features.CELL_ID])[[
                Axes.X.value, Axes.Y.value, Axes.ZPLANE.value]]
        min_ = grouped.min()
        max_ = grouped.max()
        coordinate_df = min_ + (max_ - min_) / 2
        metadata = {name: (Features.CELLS, data.values) for name, data in coordinate_df.items()}
        metadata[Features.AREA] = (Features.CELLS, np.full(counts.shape[0], fill_value=np.nan))
        # add genes to the metadata
        metadata.update({Features.GENES: counts.columns.values})
        metadata.update({Features.CELL_ID: (Features.CELLS, counts.index.values)})

        mat = ExpressionMatrix(
            data=counts.values,
            dims=(Features.CELLS, Features.GENES),
            coords=metadata,
            name='expression_matrix'
        )
        return mat



from starfish.core.codebook.codebook import Codebook
from starfish.core.intensity_table.decoded_intensity_table import DecodedIntensityTable
from starfish.core.intensity_table.intensity_table_coordinates import \
    transfer_physical_coords_to_intensity_table
from starfish.core.spots.DecodeSpots.trace_builders import TRACE_BUILDERS
from starfish.core.types import Number, SpotFindingResults, TraceBuildingStrategies
from ._base import DecodeSpotsAlgorithm


class MetricDistance(DecodeSpotsAlgorithm):
    """
    Normalizes both the magnitudes of the codes and the spot intensities, then decodes spots by
    assigning each spot to the closest code, measured by the provided metric.

    Codes greater than max_distance from the nearest code, or dimmer than min_intensity, are
    discarded.

    Parameters
    ----------
    codebook : Codebook
        codebook containing targets the experiment was designed to quantify
    max_distance : Number
        spots greater than this distance from their nearest target are not decoded
    min_intensity : Number
        spots dimmer than this intensity are not decoded
    metric : str
        the metric to use to measure distance. Can be any metric that satisfies the triangle
        inequality that is implemented by :py:mod:`scipy.spatial.distance` (default "euclidean")
    norm_order : int
        the norm to use to normalize the magnitudes of spots and codes (default 2, L2 norm)
    trace_building_strategy: TraceBuildingStrategies
        Defines the strategy for building spot traces to decode across rounds and chs of spot
        finding results.
    anchor_round : Optional[int]
        Only applicable if trace_building_strategy is TraceBuildingStrategies.NEAREST_NEIGHBORS.
        The imaging round against which other rounds will be checked for spots in the same
        approximate pixel location.
    search_radius : Optional[int]
        Only applicable if trace_building_strategy is TraceBuildingStrategies.NEAREST_NEIGHBORS.
        Number of pixels over which to search for spots in other rounds and channels.
    """

    def __init__(
        self,
        codebook: Codebook,
        max_distance: Number,
        min_intensity: Number,
        norm_order: int = 2,
        metric: str = "euclidean",
        trace_building_strategy: TraceBuildingStrategies = TraceBuildingStrategies.EXACT_MATCH,
        anchor_round: int = 1,
        search_radius: int = 3,
    ) -> None:
        self.codebook = codebook
        self.max_distance = max_distance
        self.min_intensity = min_intensity
        self.norm_order = norm_order
        self.metric = metric
        self.trace_builder = TRACE_BUILDERS[trace_building_strategy]
        self.anchor_round = anchor_round
        self.search_radius = search_radius

    def run(
        self,
        spots: SpotFindingResults,
        *args
    ) -> DecodedIntensityTable:
        """Decode spots by selecting the max-valued channel in each sequencing round

        Parameters
        ----------
        spots : SpotFindingResults
            A Dict of tile indices and their corresponding measured spots

        Returns
        -------
        DecodedIntensityTable :
            IntensityTable decoded and appended with Features.TARGET and Features.QUALITY values.

        """

        intensities = self.trace_builder(spot_results=spots,
                                         anchor_round=self.anchor_round,
                                         search_radius=self.search_radius)
        transfer_physical_coords_to_intensity_table(intensity_table=intensities, spots=spots)
        return self.codebook.decode_metric(
            intensities,
            max_distance=self.max_distance,
            min_intensity=self.min_intensity,
            norm_order=self.norm_order,
            metric=self.metric,
        )



import warnings
from typing import Optional, Union

from starfish.core.morphology.binary_mask import BinaryMaskCollection
from starfish.core.types import FunctionSource, FunctionSourceBundle
from ._base import FilterAlgorithm


class Map(FilterAlgorithm):
    """
    Map from input to output by applying a specified function to the input.  The output must have
    the same shape as the input.

    Parameters
    ----------
    func : Union[str, FunctionSourceBundle]
        Function to apply across to each of the tiles in the input.

        If this value is a string, then the ``module`` parameter is consulted to determine which
        python package is used to find the function.  If ``module`` is not specified, then the
        default is :py:attr:`FunctionSource.np`.

        If this value is a ``FunctionSourceBundle``, then the python package and module name is
        obtained from the bundle.
    module : Optional[FunctionSource]
        Python module that serves as the source of the function.  It must be listed as one of the
        members of :py:class:`FunctionSource`.

        Currently, the supported FunctionSources are:
        - ``np``: the top-level package of numpy
        - ``scipy``: the top-level package of scipy

        This is being deprecated in favor of specifying the function as a ``FunctionSourceBundle``.
    Examples
    --------
    Applying a binary opening function.
        >>> from starfish.core.morphology.binary_mask.test import factories
        >>> from starfish.morphology import Filter
        >>> from starfish.types import FunctionSource
        >>> from skimage.morphology import disk
        >>> binary_mask_collection = factories.binary_mask_collection_2d()
        >>> opener = Filter.Map(FunctionSource.scipy("morphology.binary_opening"), disk(4))
        >>> opened = opener.run(binary_mask_collection)
    """

    def __init__(
        self,
            func: Union[str, FunctionSourceBundle],
            *func_args,
            module: FunctionSource = FunctionSource.np,
            **func_kwargs,
    ) -> None:
        if isinstance(func, str):
            if module is not None:
                warnings.warn(
                    f"The module parameter is being deprecated.  Use "
                    f"`func=FunctionSource.{module.name}{func} instead.",
                    DeprecationWarning)
            else:
                module = FunctionSource.np
            self._func = module(func)
        elif isinstance(func, FunctionSourceBundle):
            if module is not None:
                raise ValueError(
                    f"When passing in the function as a `FunctionSourceBundle`, module should not "
                    f"be set."
                )
            self._func = func
        self._func_args = func_args
        self._func_kwargs = func_kwargs

    def run(
            self,
            binary_mask_collection: BinaryMaskCollection,
            n_processes: Optional[int] = None,
            *args,
            **kwargs
    ) -> BinaryMaskCollection:
        """Map from input to output by applying a specified function to the input.

        Parameters
        ----------
        binary_mask_collection : BinaryMaskCollection
            BinaryMaskCollection to be filtered.
        n_processes : Optional[int]
            The number of processes to use for apply. If None, uses the output of os.cpu_count()
            (default = None).

        Returns
        -------
        BinaryMaskCollection
            Return the results of filter as a new BinaryMaskCollection.
        """

        # Apply the reducing function
        return binary_mask_collection._apply(
            self._func.resolve(),
            *self._func_args,
            **self._func_kwargs)



import importlib
from dataclasses import dataclass
from enum import Enum
from typing import (
    Callable,
    cast,
    Mapping,
    Optional,
)


@dataclass
class FunctionSourceBundle:
    """The combination of a ``FunctionSource`` and the method we are looking for."""
    top_level_package: str
    requested_method: str
    actual_method: str

    def resolve(self) -> Callable:
        """Resolve a method.  The method itself might be enclosed in a package, such as
        subpackage.actual_method.  In that case, we will need to attempt to resolve it in the
        following sequence:

        1. import top_level_package, then try to resolve subpackage.actual_method recursively
           through ``getattr`` calls.
        2. import top_level_package.subpackage, then try to resolve actual_method through
           ``gettatr`` calls.

        This is done instead of just creating a bunch of FunctionSource for libraries that have
        a lot of packages that are not implicitly imported by importing the top-level package.
        """
        method_splitted = self.actual_method.split(".")
        splitted = [self.top_level_package]
        splitted.extend(method_splitted)

        for divider in range(1, len(splitted)):
            import_section = splitted[:divider]
            getattr_section = splitted[divider:]

            imported = importlib.import_module(".".join(import_section))

            try:
                for getattr_name in getattr_section:
                    imported = getattr(imported, getattr_name)
                return cast(Callable, imported)
            except AttributeError:
                pass

        raise AttributeError(
            f"Unable to resolve the method {self.requested_method} from package "
            f"{self.top_level_package}")


class FunctionSource(Enum):
    """Each FunctionSource declares a package from which reduction methods can be obtained.
    Generally, the packages should be those that are included as starfish's dependencies for
    reproducibility.

    Many packages are broken into subpackages which are not necessarily implicitly imported when
    importing the top-level package.  For example, ``scipy.linalg`` is not implicitly imported
    when one imports ``scipy``.  To avoid the complexity of enumerating each scipy subpackage in
    FunctionSource, we assemble the fully-qualified method name, and then try all the
    permutations of how one could import that method.

    In the example of ``scipy.linalg.norm``, we try the following:

    1. import ``scipy``, attempt to resolve ``linalg.norm``.
    2. import ``scipy.linalg``, attempt to resolve ``norm``.
    """

    def __init__(self, top_level_package: str, aliases: Optional[Mapping[str, str]] = None):
        self.top_level_package = top_level_package
        self.aliases = aliases or {}

    def __call__(self, method: str) -> FunctionSourceBundle:
        return FunctionSourceBundle(
            self.top_level_package, method, self.aliases.get(method, method))

    np = ("numpy", {'max': 'amax'})
    """Function source for the numpy libraries"""
    scipy = ("scipy",)
    skimage = ("skimage",)



from functools import partial
from typing import Optional

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from ._base import FilterAlgorithm
from .util import determine_axes_to_group_by


class ClipPercentileToZero(FilterAlgorithm):
    """
    Image clipping filter that clips values below a minimum percentile and
    above a maximum percentile, and follows up by subtracting the minimum
    percentile value from the image.

    By default, these min and max percentiles are set to 0 and 100
    respectively, which will result in the filter doing nothing.

    This is a wrapper for :py:func:`numpy.clip`.

    Parameters
    ----------
    p_min : int
        Values below this percentile are set to p_min, and the p_min value
        is subtracted from the image (default 0)
    p_max : int
        Values above this percentile are set to p_max (default 100)
    min_coeff : float
        Apply a coefficient to the minimum percentile value. (default 1.0)
    max_coeff : float
        Apply a coefficient to the maximum percentile value. (default 1.0)
    is_volume : bool
        If True, 3D (z, y, x) volumes will be filtered. By default, filter 2D
        (y, x) tiles
    """

    def __init__(
        self, p_min: int = 0, p_max: int = 100,
        min_coeff: float = 1.0, max_coeff: float = 1.0,
        is_volume: bool = False,
    ) -> None:

        self.p_min = p_min
        self.p_max = p_max
        self.min_coeff = min_coeff
        self.max_coeff = max_coeff
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"p_min": 0, "p_max": 100,
                                   "min_coeff": 1.0, "max_coeff": 1.0}

    @staticmethod
    def _clip_percentile_to_zero(
            image: xr.DataArray,
            p_min: int,
            p_max: int,
            min_coeff: float,
            max_coeff: float) -> xr.DataArray:
        v_min, v_max = np.percentile(image, [p_min, p_max])
        v_min = min_coeff * v_min
        v_max = max_coeff * v_max
        return image.clip(min=v_min, max=v_max) - np.float32(v_min)

    def run(self, stack: ImageStack, in_place: bool = False,
            verbose: bool = False, n_processes: Optional[int] = None,
            *args) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter.
            If None, defaults to the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.
            Otherwise return the original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        clip_percentile_to_zero = partial(
            self._clip_percentile_to_zero,
            p_min=self.p_min, p_max=self.p_max,
            min_coeff=self.min_coeff, max_coeff=self.max_coeff
        )
        result = stack.apply(
            clip_percentile_to_zero,
            group_by=group_by, verbose=verbose,
            in_place=in_place, n_processes=n_processes
        )
        return result
